# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test arbitrary nesting of bit field types."""

# pylint: disable=unexpected-keyword-arg

from baseline import Baseline

from plum.bitfields import BitFields, bitfield


class Inner(BitFields, nested=True, default=0x4):

    """Sample datatype for innermost part of structure."""

    il0: int = bitfield(lsb=0, size=1, ignore=True)
    il1: int = bitfield(lsb=1, size=1, default=1)
    # empty bitfield(lsb=2, size=1)
    il2: int = bitfield(lsb=3, size=2)


class Middle(BitFields, nested=True, default=0xC, ignore=0x4):

    """Sample datatype for middle part of structure."""

    ml0: int = bitfield(lsb=0, size=1, ignore=True)
    ml1: int = bitfield(lsb=1, size=1, default=1)
    # empty bitfield(lsb=2, size=2)
    ml2: Inner = bitfield(lsb=4, size=5, typ=Inner)


class Outer(BitFields, nbytes=2, default=0x70, ignore=0x10):

    """Sample datatype for outermost part of structure."""

    ol0: int = bitfield(lsb=0, size=2, ignore=True)
    ol1: int = bitfield(lsb=2, size=2, default=1)
    # empty bitfield(lsb=4, size=3)
    ol2: Middle = bitfield(lsb=7, size=9, typ=Middle)


class TestNested:

    """Verify bit fields may be nested within bitfields arbitrarily.

    Following aspects covered without need for dedicated test cases:

        - dump
        - instantiating with class instances

    """

    expected_dump = Baseline(
        """
        +----------+---------+-------+-------------------+--------------------+
        | Offset   | Access  | Value | Bytes             | Format             |
        +----------+---------+-------+-------------------+--------------------+
        | 0        |         | 46964 | 74 b7             | Outer (BitFields)  |
        |  [0:2]   | ol0     | 0     | ........ ......00 | int                |
        |  [2:4]   | ol1     | 1     | ........ ....01.. | int                |
        |          | ol2     |       |                   | Middle (BitFields) |
        |  [7]     |   ml0   | 0     | ........ 0....... | int                |
        |  [8]     |   ml1   | 1     | .......1 ........ | int                |
        |          |   ml2   |       |                   | Inner (BitFields)  |
        |  [11]    |     il0 | 0     | ....0... ........ | int                |
        |  [12]    |     il1 | 1     | ...1.... ........ | int                |
        |  [14:16] |     il2 | 2     | 10...... ........ | int                |
        +----------+---------+-------+-------------------+--------------------+
        """
    )

    def test_get(self):
        """Verify read access."""
        instance = Outer(
            ol0=0, ol1=1, ol2=Middle(ml0=0, ml1=1, ml2=Inner(il0=0, il1=1, il2=2))
        )

        assert instance == 0xB774

        assert instance.ol0 == 0
        assert instance.ol1 == 1
        assert instance.ol2 == 0x16E

        assert instance.ol2.ml0 == 0
        assert instance.ol2.ml1 == 1
        assert instance.ol2.ml2 == 0x16

        assert instance.ol2.ml2.il0 == 0
        assert instance.ol2.ml2.il1 == 1
        assert instance.ol2.ml2.il2 == 2

        assert str(instance.dump) == self.expected_dump

    def test_set(self):
        """Verify write access."""
        # initialize all fields to zero (except initialize to 1 all fields that
        # should be 0)
        instance = Outer(
            ol0=1, ol1=0, ol2=Middle(ml0=1, ml1=0, ml2=Inner(il0=1, il1=0, il2=0))
        )

        # fill in each field with expected value
        instance.ol0 = 0
        instance.ol1 = 1
        instance.ol2.ml0 = 0
        instance.ol2.ml1 = 1
        instance.ol2.ml2.il0 = 0
        instance.ol2.ml2.il1 = 1
        instance.ol2.ml2.il2 = 2

        assert str(instance.dump) == self.expected_dump

    def test_ignore_fields(self):
        """Verify ignored fields at lower level don't cause mis-compare.

        Verify bitfield(..., ignore=True) has impact from nested levels.

        """
        item1 = Outer(
            ol0=0, ol1=1, ol2=Middle(ml0=0, ml1=1, ml2=Inner(il0=0, il1=1, il2=2))
        )
        item2 = Outer(
            ol0=3, ol1=1, ol2=Middle(ml0=1, ml1=1, ml2=Inner(il0=1, il1=1, il2=2))
        )
        assert item1 == item2

    def test_ignore_background(self):
        """Verify ignored background bits at lower level don't cause mis-compare.

        Verify Middle(..., ignore=0x4) has impact.

        """
        assert Outer.from_int(0x210) == Outer.from_int(0)

    def test_fill(self):
        """Verify fill patterns for each class are meshed properly."""

        expected_dump = Baseline(
            """
            +----------+---------+-------+-------------------+--------------------+
            | Offset   | Access  | Value | Bytes             | Format             |
            +----------+---------+-------+-------------------+--------------------+
            | 0        |         | 9840  | 70 26             | Outer (BitFields)  |
            |  [0:2]   | ol0     | 0     | ........ ......00 | int                |
            |  [2:4]   | ol1     | 0     | ........ ....00.. | int                |
            |          | ol2     |       |                   | Middle (BitFields) |
            |  [7]     |   ml0   | 0     | ........ 0....... | int                |
            |  [8]     |   ml1   | 0     | .......0 ........ | int                |
            |          |   ml2   |       |                   | Inner (BitFields)  |
            |  [11]    |     il0 | 0     | ....0... ........ | int                |
            |  [12]    |     il1 | 0     | ...0.... ........ | int                |
            |  [14:16] |     il2 | 0     | 00...... ........ | int                |
            +----------+---------+-------+-------------------+--------------------+
            """
        )

        instance = Outer(
            ol0=0, ol1=0, ol2=Middle(ml0=0, ml1=0, ml2=Inner(il0=0, il1=0, il2=0))
        )

        assert str(instance.dump) == expected_dump

    def test_repr(self):
        """Test representation."""
        instance = Outer(
            ol0=0, ol1=0, ol2=Middle(ml0=0, ml1=0, ml2=Inner(il0=0, il1=0, il2=0))
        )

        expected_repr = Baseline(
            """
            Outer(ol0=0, ol1=0, ol2=Middle(ml0=0, ml1=0, ml2=Inner(il0=0, il1=0, il2=0)))
            """
        )

        assert repr(instance) == expected_repr
        assert str(instance) == expected_repr


class TestGetterOptimization:

    """Test elimination of lsb added to bit offset in getter."""

    def test_getter(self):
        class Nested(BitFields, nested=True, fieldorder="least_to_most"):
            n1: int = bitfield(size=2)
            n2: int = bitfield(size=2)

        class Outside(BitFields, fieldorder="least_to_most"):
            o1: Nested = bitfield(size=4, typ=Nested)
            o2: int = bitfield(size=4)

        bitfields = Outside.from_int(3)

        assert bitfields.o1.n1 == 3



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test BitFields data store transform properties."""
# pylint: disable=comparison-with-callable

from typing import Union
from enum import IntEnum

from plum.bitfields import BitFields, bitfield
from plum.enum import EnumX


class TestDefault:

    """Test with as many left to default as possible."""

    class Sample(BitFields):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1)
        m1: int = bitfield(size=1)

    def test_nbytes(self):
        assert self.Sample.nbytes == 1
        assert self.Sample.from_int(0).nbytes == 1

    def test_byteorder(self):
        assert self.Sample.byteorder == "little"

    def test_default(self):
        assert self.Sample.default == 0

    def test_ignore(self):
        assert self.Sample.ignore == 0

    def test_nested(self):
        assert self.Sample.nested is False


class TestKeyword:

    """Test explicitly defined with keyword argument."""

    class Sample(
        BitFields, nbytes=2, byteorder="big", default=1, ignore=2, nested=True
    ):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1)
        m1: int = bitfield(size=1)

    def test_nbytes(self):
        assert self.Sample.nbytes == 2
        assert self.Sample.from_int(0).nbytes == 2

    def test_byteorder(self):
        assert self.Sample.byteorder == "big"

    def test_default(self):
        assert self.Sample.default == 1

    def test_ignore(self):
        assert self.Sample.ignore == 2

    def test_nested(self):
        assert self.Sample.nested is True


class TestNameAndHint:
    class Sample(BitFields):

        m0: int = bitfield(size=1)
        m1: int = bitfield(size=1)

    def test_name(self):
        assert self.Sample.name == "Sample (BitFields)"
        assert self.Sample.__name__ == "Sample"

    def test_hint(self):
        assert self.Sample.__hint__ == "Sample"
        assert self.Sample.__hint__ == "Sample"


class Bits(IntEnum):

    ZEROES = 0x00
    ONES = 0x03


class TestBitFieldAccess:
    class Sample(BitFields):

        number: int = bitfield(size=3, lsb=0)
        boolean: int = bitfield(size=1, lsb=3, typ=bool)
        enumeration: Bits = bitfield(size=2, lsb=4, typ=Bits)
        enumeration_transform: Union[Bits, int] = bitfield(
            size=2, lsb=6, typ=EnumX(enum=Bits, strict=False)
        )

    def test_get(self):
        assert self.Sample.from_int(0xF8).number == 0
        assert self.Sample.from_int(0x07).number == 7

        assert self.Sample.from_int(0xF7).boolean is False
        assert self.Sample.from_int(0x08).boolean is True

        assert self.Sample.from_int(0xCF).enumeration is Bits.ZEROES
        assert self.Sample.from_int(0x30).enumeration is Bits.ONES

        assert self.Sample.from_int(0x3F).enumeration_transform is Bits.ZEROES
        assert self.Sample.from_int(0xC0).enumeration_transform is Bits.ONES

    def test_set(self):
        sample = self.Sample.from_int(0)
        sample.number = 7
        assert int(sample) == 0x07

        sample = self.Sample.from_int(0)
        sample.boolean = True
        assert int(sample) == 0x08

        sample = self.Sample.from_int(0)
        sample.enumeration = Bits.ONES
        assert int(sample) == 0x30

        sample = self.Sample.from_int(0)
        sample.enumeration_transform = Bits.ONES
        assert int(sample) == 0xC0



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test float transform properties."""

from plum.float import FloatX


class TestDefault:

    """Test with as many left to default as possible."""

    float32 = FloatX(nbytes=4)

    def test_half(self):
        xform = FloatX(nbytes=2)
        assert xform.__hint__ == "float"
        assert xform.name == "half float"
        assert xform.nbytes == 2
        assert xform.byteorder == "little"

    def test_single(self):
        xform = FloatX(nbytes=4)
        assert xform.__hint__ == "float"
        assert xform.name == "single float"
        assert xform.nbytes == 4
        assert xform.byteorder == "little"

    def test_double(self):
        xform = FloatX(nbytes=8)
        assert xform.__hint__ == "float"
        assert xform.name == "double float"
        assert xform.nbytes == 8
        assert xform.byteorder == "little"


class TestPositional:

    """Test explicitly defined with positional argument."""

    float32 = FloatX(4, "big", "name")

    def test_hint(self):
        assert self.float32.__hint__ == "float"

    def test_name(self):
        assert self.float32.name == "name"

    def test_nbytes(self):
        assert self.float32.nbytes == 4

    def test_byteorder(self):
        assert self.float32.byteorder == "big"


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    float32 = FloatX(nbytes=4, byteorder="big", name="name")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test sized_member() member definitions."""

# pylint: disable=too-many-lines

import sys

import pytest
from baseline import Baseline

from plum.array import ArrayX
from plum.conformance import Case, CaseData, wrap_message
from plum.exceptions import UnpackError
from plum.littleendian import uint8, uint16
from plum.str import StrX
from plum.structure import Structure, member, sized_member
from plum.utilities import unpack

ascii_greedy = StrX(name="ascii", encoding="ascii")


class Simple(Structure):

    """Sized string."""

    size: int = member(fmt=uint8, compute=True)
    string: str = sized_member(size=size, fmt=ascii_greedy)
    bookend: int = member(fmt=uint8)


sizedstrarray = ArrayX(name="sizedstrarray", fmt=Simple)


class TestSimple(Case):

    data = CaseData(
        fmt=Simple,
        bindata=b"\x0cHello World!\x99",
        nbytes=None,
        values=(
            Simple(size=12, string="Hello World!", bookend=0x99),
            Simple(string="Hello World!", bookend=0x99),
        ),
        dump=Baseline(
            """
            +--------+----------+----------------+-------------------------------------+--------------------+
            | Offset | Access   | Value          | Bytes                               | Format             |
            +--------+----------+----------------+-------------------------------------+--------------------+
            |        |          |                |                                     | Simple (Structure) |
            |  0     | size     | 12             | 0c                                  | uint8              |
            |        | string   |                |                                     | ascii              |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
            | 13     | bookend  | 153            | 99                                  | uint8              |
            +--------+----------+----------------+-------------------------------------+--------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+----------+----------------+-------------------------------------+--------------------+
            | Offset | Access   | Value          | Bytes                               | Format             |
            +--------+----------+----------------+-------------------------------------+--------------------+
            |        |          |                |                                     | Simple (Structure) |
            |  0     | size     | 12             | 0c                                  | uint8              |
            |        | string   |                |                                     | ascii              |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
            | 13     | bookend  | 153            | 99                                  | uint8              |
            +--------+----------+----------------+-------------------------------------+--------------------+
            | 14     |          | <excess bytes> | 99                                  |                    |
            +--------+----------+----------------+-------------------------------------+--------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
        +--------+----------+----------------------+-------------------------------------+--------------------+
        | Offset | Access   | Value                | Bytes                               | Format             |
        +--------+----------+----------------------+-------------------------------------+--------------------+
        |        |          |                      |                                     | Simple (Structure) |
        |  0     | size     | 12                   | 0c                                  | uint8              |
        |        | string   |                      |                                     | ascii              |
        |  1     |   [0:12] | 'Hello World!'       | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
        |        | bookend  | <insufficient bytes> |                                     | uint8              |
        +--------+----------+----------------------+-------------------------------------+--------------------+

        InsufficientMemoryError occurred during unpack operation:

        1 too few bytes to unpack uint8, 1 needed, only 0 available
        """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, size: Optional[int] = None, string: str, bookend: int) -> None:
                self[:] = (size, string, bookend)

            @size.getter
            def size(self) -> int:
                if self[0] is None:
                    self[0] = self.unpack(self.ipack())[0]

                return self[0]

            @size.setter
            def size(self, value: int) -> None:
                self[0] = value

            @string.getter
            def string(self) -> str:
                return self[1]

            @string.setter
            def string(self, value: str) -> None:
                self[1] = value
                self[0] = None  # re-compute 'size' member

            @bookend.getter
            def bookend(self) -> int:
                return self[2]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_size, m_string, m_bookend) = value

                if dump is None:
                    size_pieces_index = len(pieces)
                    if m_size is None:
                        pieces.append(b"")
                    else:
                        cls.size.fmt.__pack__(m_size, pieces, dump)

                    if m_size is None:
                        string_pieces_index = len(pieces)
                        cls.string.fmt.__pack__(m_string, pieces, dump)
                        m_size = len(b"".join(pieces[string_pieces_index:]))
                        cls.size.fmt.__pack__(m_size, pieces, dump)
                        pieces[size_pieces_index] = pieces.pop()
                    else:
                        cls.string.fmt.__pack__(m_string, pieces, dump)

                    cls.bookend.fmt.__pack__(m_bookend, pieces, dump)

                else:
                    size_dump = dump.add_record(access="size", fmt=cls.size.fmt)
                    size_pieces_index = len(pieces)
                    if m_size is None:
                        size_dump.value = "<skipped>"
                        pieces.append(b"")
                    else:
                        cls.size.fmt.__pack__(m_size, pieces, size_dump)

                    string_dump = dump.add_record(access="string", fmt=cls.string.fmt)
                    if m_size is None:
                        string_pieces_index = len(pieces)
                        cls.string.fmt.__pack__(m_string, pieces, string_dump)
                        m_size = len(b"".join(pieces[string_pieces_index:]))
                        cls.size.fmt.__pack__(m_size, pieces, size_dump)
                        pieces[size_pieces_index] = pieces.pop()
                    else:
                        cls.string.fmt.__pack__(m_string, pieces, string_dump)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    cls.bookend.fmt.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Simple", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_size, offset = cls.size.fmt.__unpack__(buffer, offset, dump)

                    buffer, string_buffer = buffer[:offset + m_size], buffer
                    if len(buffer) < offset + m_size:
                        cls.string.report_insufficient_bytes(buffer, offset, m_size)
                    m_string, offset = cls.string.fmt.__unpack__(buffer, offset, dump)
                    if offset < len(buffer):
                        cls.string.report_extra_bytes(buffer[offset:])
                    buffer = string_buffer

                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, dump)

                else:
                    size_dump = dump.add_record(access="size", fmt=cls.size.fmt)
                    m_size, offset = cls.size.fmt.__unpack__(buffer, offset, size_dump)

                    string_dump = dump.add_record(access="string", fmt=cls.string.fmt)
                    buffer, string_buffer = buffer[:offset + m_size], buffer
                    if len(buffer) < offset + m_size:
                        cls.string.report_insufficient_bytes(buffer, offset, m_size, string_dump)
                    m_string, offset = cls.string.fmt.__unpack__(buffer, offset, string_dump)
                    if offset < len(buffer):
                        cls.string.report_extra_bytes(buffer[offset:], string_dump)
                    buffer = string_buffer

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, bookend_dump)

                structure[:] = (m_size, m_string, m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_size, _s_string, _s_bookend = self
                    if _s_size is None:
                        _s_size, _s_string, _s_bookend = self.unpack(self.ipack())
                    _o_size, _o_string, _o_bookend = other
                    if _o_size is None:
                        _o_size, _o_string, _o_bookend = self.unpack(other.ipack())
                    return (_s_size, _s_string, _s_bookend) == (_o_size, _o_string, _o_bookend)
                else:
                    return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_size, _s_string, _s_bookend = self
                    if _s_size is None:
                        _s_size, _s_string, _s_bookend = self.unpack(self.ipack())
                    _o_size, _o_string, _o_bookend = other
                    if _o_size is None:
                        _o_size, _o_string, _o_bookend = self.unpack(other.ipack())
                    return (_s_size, _s_string, _s_bookend) != (_o_size, _o_string, _o_bookend)
                else:
                    return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(size={self.size!r}, string={self.string!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )

    def test_pack_explicit_size(self):
        """Test explicit size value used instead of computed size."""
        expected_dump = Baseline(
            """
            +--------+----------+----------------+-------------------------------------+--------------------+
            | Offset | Access   | Value          | Bytes                               | Format             |
            +--------+----------+----------------+-------------------------------------+--------------------+
            |        |          |                |                                     | Simple (Structure) |
            |  0     | size     | 32             | 20                                  | uint8              |
            |        | string   |                |                                     | ascii              |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
            | 13     | bookend  | 153            | 99                                  | uint8              |
            +--------+----------+----------------+-------------------------------------+--------------------+
            """
        )

        buffer, dump = Simple(
            size=32, string="Hello World!", bookend=0x99
        ).ipack_and_dump()
        assert str(dump) == expected_dump
        assert buffer == b"\x20Hello World!\x99"


class TestMetaException:
    def test_bad_reference(self):
        """Verify exception/message if sized_member gets invalid size member."""
        with pytest.raises(TypeError) as trap:

            class Bad(Structure):  # pylint: disable=unused-variable

                size: int = member(fmt=uint16, compute=True)
                array: list = sized_member(size="junk", fmt=sizedstrarray)

        expected = Baseline(
            """
            invalid 'size', must be a 'member()'
            """
        )

        assert wrap_message(trap.value) == expected

    def test_missing_association(self):
        """Verify exception/message if sized_member definition missing."""
        with pytest.raises(TypeError) as trap:

            class Bad(Structure):  # pylint: disable=unused-variable

                size: int = member(fmt=uint16, compute=True)

        expected = Baseline(
            """
            'size' member never associated with member used to compute it
            """
        )

        assert wrap_message(trap.value) == expected

    def test_excess_bytes(self):
        """Test outer limit more restrictive than inner limit (string case)."""

        class ExcessBytes(Structure):
            """Excess Bytes string."""

            size: int = member(fmt=uint8, compute=True)
            fixed: int = sized_member(size=size, fmt=uint8)

        memory_bytes = b"\x20" + bytes(32)

        with pytest.raises(UnpackError) as trap:
            unpack(ExcessBytes, memory_bytes)

        expected = Baseline(
            """
            +--------+--------+----------------+-------------------------------------------------+-------------------------+
            | Offset | Access | Value          | Bytes                                           | Format                  |
            +--------+--------+----------------+-------------------------------------------------+-------------------------+
            |        |        |                |                                                 | ExcessBytes (Structure) |
            |  0     | size   | 32             | 20                                              | uint8                   |
            |  1     | fixed  | 0              | 00                                              | uint8                   |
            +--------+--------+----------------+-------------------------------------------------+-------------------------+
            |  2     |        | <excess bytes> | 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 |                         |
            | 18     |        |                | 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00    |                         |
            +--------+--------+----------------+-------------------------------------------------+-------------------------+

            ExcessMemoryError occurred during unpack operation:

            31 unconsumed bytes
            """
        )

        assert wrap_message(trap.value) == expected


class Ratio(Structure):

    """Sized string with ratio usage."""

    size: int = member(fmt=uint8, compute=True)
    string: str = sized_member(size=size, fmt=ascii_greedy, ratio=2)
    bookend: int = member(fmt=uint8)


RATIO_DUMP = Baseline(
    """
    +--------+-------------+----------------+-------------------------------------+--------+
    | Offset | Access      | Value          | Bytes                               | Format |
    +--------+-------------+----------------+-------------------------------------+--------+
    |        |             |                |                                     | Ratio  |
    |  0     | size [0]    | 6              | 06                                  | uint8  |
    |        | string [1]  |                |                                     | ascii  |
    |  1     |   [0:12]    | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |        |
    | 13     | bookend [2] | 153            | 99                                  | uint8  |
    +--------+-------------+----------------+-------------------------------------+--------+
    """
)


class TestRatio(Case):

    """Test ratio parameter usage."""

    data = CaseData(
        fmt=Ratio,
        bindata=b"\x06Hello World!\x99",
        nbytes=None,
        values=(
            Ratio(size=6, string="Hello World!", bookend=0x99),
            Ratio(string="Hello World!", bookend=0x99),
        ),
        dump=Baseline(
            """
            +--------+----------+----------------+-------------------------------------+-------------------+
            | Offset | Access   | Value          | Bytes                               | Format            |
            +--------+----------+----------------+-------------------------------------+-------------------+
            |        |          |                |                                     | Ratio (Structure) |
            |  0     | size     | 6              | 06                                  | uint8             |
            |        | string   |                |                                     | ascii             |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                   |
            | 13     | bookend  | 153            | 99                                  | uint8             |
            +--------+----------+----------------+-------------------------------------+-------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+----------+----------------+-------------------------------------+-------------------+
            | Offset | Access   | Value          | Bytes                               | Format            |
            +--------+----------+----------------+-------------------------------------+-------------------+
            |        |          |                |                                     | Ratio (Structure) |
            |  0     | size     | 6              | 06                                  | uint8             |
            |        | string   |                |                                     | ascii             |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                   |
            | 13     | bookend  | 153            | 99                                  | uint8             |
            +--------+----------+----------------+-------------------------------------+-------------------+
            | 14     |          | <excess bytes> | 99                                  |                   |
            +--------+----------+----------------+-------------------------------------+-------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
        +--------+----------+----------------------+-------------------------------------+-------------------+
        | Offset | Access   | Value                | Bytes                               | Format            |
        +--------+----------+----------------------+-------------------------------------+-------------------+
        |        |          |                      |                                     | Ratio (Structure) |
        |  0     | size     | 6                    | 06                                  | uint8             |
        |        | string   |                      |                                     | ascii             |
        |  1     |   [0:12] | 'Hello World!'       | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                   |
        |        | bookend  | <insufficient bytes> |                                     | uint8             |
        +--------+----------+----------------------+-------------------------------------+-------------------+

        InsufficientMemoryError occurred during unpack operation:

        1 too few bytes to unpack uint8, 1 needed, only 0 available
        """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, size: Optional[int] = None, string: str, bookend: int) -> None:
                self[:] = (size, string, bookend)

            @size.getter
            def size(self) -> int:
                if self[0] is None:
                    self[0] = self.unpack(self.ipack())[0]

                return self[0]

            @size.setter
            def size(self, value: int) -> None:
                self[0] = value

            @string.getter
            def string(self) -> str:
                return self[1]

            @string.setter
            def string(self, value: str) -> None:
                self[1] = value
                self[0] = None  # re-compute 'size' member

            @bookend.getter
            def bookend(self) -> int:
                return self[2]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_size, m_string, m_bookend) = value

                if dump is None:
                    size_pieces_index = len(pieces)
                    if m_size is None:
                        pieces.append(b"")
                    else:
                        cls.size.fmt.__pack__(m_size, pieces, dump)

                    if m_size is None:
                        string_pieces_index = len(pieces)
                        cls.string.fmt.__pack__(m_string, pieces, dump)
                        m_size = int(len(b"".join(pieces[string_pieces_index:])) // 2)
                        cls.size.fmt.__pack__(m_size, pieces, dump)
                        pieces[size_pieces_index] = pieces.pop()
                    else:
                        cls.string.fmt.__pack__(m_string, pieces, dump)

                    cls.bookend.fmt.__pack__(m_bookend, pieces, dump)

                else:
                    size_dump = dump.add_record(access="size", fmt=cls.size.fmt)
                    size_pieces_index = len(pieces)
                    if m_size is None:
                        size_dump.value = "<skipped>"
                        pieces.append(b"")
                    else:
                        cls.size.fmt.__pack__(m_size, pieces, size_dump)

                    string_dump = dump.add_record(access="string", fmt=cls.string.fmt)
                    if m_size is None:
                        string_pieces_index = len(pieces)
                        cls.string.fmt.__pack__(m_string, pieces, string_dump)
                        m_size = int(len(b"".join(pieces[string_pieces_index:])) // 2)
                        cls.size.fmt.__pack__(m_size, pieces, size_dump)
                        pieces[size_pieces_index] = pieces.pop()
                    else:
                        cls.string.fmt.__pack__(m_string, pieces, string_dump)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    cls.bookend.fmt.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Ratio", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_size, offset = cls.size.fmt.__unpack__(buffer, offset, dump)

                    string_nbytes = int(m_size * 2)
                    buffer, string_buffer = buffer[:offset + string_nbytes], buffer
                    if len(buffer) < offset + string_nbytes:
                        cls.string.report_insufficient_bytes(buffer, offset, string_nbytes)
                    m_string, offset = cls.string.fmt.__unpack__(buffer, offset, dump)
                    if offset < len(buffer):
                        cls.string.report_extra_bytes(buffer[offset:])
                    buffer = string_buffer

                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, dump)

                else:
                    size_dump = dump.add_record(access="size", fmt=cls.size.fmt)
                    m_size, offset = cls.size.fmt.__unpack__(buffer, offset, size_dump)

                    string_dump = dump.add_record(access="string", fmt=cls.string.fmt)
                    string_nbytes = int(m_size * 2)
                    buffer, string_buffer = buffer[:offset + string_nbytes], buffer
                    if len(buffer) < offset + string_nbytes:
                        cls.string.report_insufficient_bytes(buffer, offset, string_nbytes, string_dump)
                    m_string, offset = cls.string.fmt.__unpack__(buffer, offset, string_dump)
                    if offset < len(buffer):
                        cls.string.report_extra_bytes(buffer[offset:], string_dump)
                    buffer = string_buffer

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, bookend_dump)

                structure[:] = (m_size, m_string, m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_size, _s_string, _s_bookend = self
                    if _s_size is None:
                        _s_size, _s_string, _s_bookend = self.unpack(self.ipack())
                    _o_size, _o_string, _o_bookend = other
                    if _o_size is None:
                        _o_size, _o_string, _o_bookend = self.unpack(other.ipack())
                    return (_s_size, _s_string, _s_bookend) == (_o_size, _o_string, _o_bookend)
                else:
                    return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_size, _s_string, _s_bookend = self
                    if _s_size is None:
                        _s_size, _s_string, _s_bookend = self.unpack(self.ipack())
                    _o_size, _o_string, _o_bookend = other
                    if _o_size is None:
                        _o_size, _o_string, _o_bookend = self.unpack(other.ipack())
                    return (_s_size, _s_string, _s_bookend) != (_o_size, _o_string, _o_bookend)
                else:
                    return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(size={self.size!r}, string={self.string!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )

    def test_pack_explicit_size(self):
        """Test explicit size value used instead of computed size."""

        expected_dump = Baseline(
            """
            +--------+----------+----------------+-------------------------------------+-------------------+
            | Offset | Access   | Value          | Bytes                               | Format            |
            +--------+----------+----------------+-------------------------------------+-------------------+
            |        |          |                |                                     | Ratio (Structure) |
            |  0     | size     | 32             | 20                                  | uint8             |
            |        | string   |                |                                     | ascii             |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                   |
            | 13     | bookend  | 153            | 99                                  | uint8             |
            +--------+----------+----------------+-------------------------------------+-------------------+
            """
        )
        buffer, dump = Ratio(
            size=32, string="Hello World!", bookend=0x99
        ).ipack_and_dump()
        assert str(dump) == expected_dump
        assert buffer == b"\x20Hello World!\x99"


class Nested(Structure):

    """Sized array (with embedded sized strings inside)."""

    size: int = member(fmt=uint8, compute=True)
    array: list = sized_member(size=size, fmt=sizedstrarray)


NESTED_DUMP = Baseline(
    """
    +--------+--------------+----------------+-------------------------------------+--------------------+
    | Offset | Access       | Value          | Bytes                               | Format             |
    +--------+--------------+----------------+-------------------------------------+--------------------+
    |        |              |                |                                     | Nested (Structure) |
    |  0     | size         | 25             | 19                                  | uint8              |
    |        | array        |                |                                     | sizedstrarray      |
    |        |   [0]        |                |                                     | Simple (Structure) |
    |  1     |     size     | 12             | 0c                                  | uint8              |
    |        |     string   |                |                                     | ascii              |
    |  2     |       [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
    | 14     |     bookend  | 153            | 99                                  | uint8              |
    |        |   [1]        |                |                                     | Simple (Structure) |
    | 15     |     size     | 9              | 09                                  | uint8              |
    |        |     string   |                |                                     | ascii              |
    | 16     |       [0:9]  | 'Good Bye!'    | 47 6f 6f 64 20 42 79 65 21          |                    |
    | 25     |     bookend  | 153            | 99                                  | uint8              |
    +--------+--------------+----------------+-------------------------------------+--------------------+
    """
)


class TestNested:

    memory_bytes = b"\x19\x0cHello World!\x99\x09Good Bye!\x99"
    array_value = [
        Simple(string="Hello World!", bookend=0x99),
        Simple(string="Good Bye!", bookend=0x99),
    ]

    def test_unpack(self):
        """Verify size member limits how much string member consumes."""
        nested = unpack(Nested, self.memory_bytes)
        assert str(nested.dump) == NESTED_DUMP  # pylint: disable=no-member

    def test_pack_implicit_size(self):
        """Verify size member computed from string."""
        buffer, dump = Nested(array=self.array_value).ipack_and_dump()
        assert str(dump) == NESTED_DUMP
        assert buffer == self.memory_bytes

    def test_pack_explicit_size(self):
        """Test explicit size value used instead of computed size."""
        expected_dump = Baseline(
            """
            +--------+--------------+----------------+-------------------------------------+--------------------+
            | Offset | Access       | Value          | Bytes                               | Format             |
            +--------+--------------+----------------+-------------------------------------+--------------------+
            |        |              |                |                                     | Nested (Structure) |
            |  0     | size         | 32             | 20                                  | uint8              |
            |        | array        |                |                                     | sizedstrarray      |
            |        |   [0]        |                |                                     | Simple (Structure) |
            |  1     |     size     | 12             | 0c                                  | uint8              |
            |        |     string   |                |                                     | ascii              |
            |  2     |       [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
            | 14     |     bookend  | 153            | 99                                  | uint8              |
            |        |   [1]        |                |                                     | Simple (Structure) |
            | 15     |     size     | 9              | 09                                  | uint8              |
            |        |     string   |                |                                     | ascii              |
            | 16     |       [0:9]  | 'Good Bye!'    | 47 6f 6f 64 20 42 79 65 21          |                    |
            | 25     |     bookend  | 153            | 99                                  | uint8              |
            +--------+--------------+----------------+-------------------------------------+--------------------+
            """
        )

        assert str(Nested(size=32, array=self.array_value).dump) == expected_dump

    def test_smaller_first_limit_array(self):
        """Test outer limit more restrictive than inner limit (array case).

        Test array special as they muck with limit.

        """
        memory_bytes = b"\x06\x0cHello World!\x99"

        with pytest.raises(UnpackError) as trap:
            unpack(Nested, memory_bytes)

        expected = Baseline(
            """
            +--------+------------+----------------------+----------------+--------------------+
            | Offset | Access     | Value                | Bytes          | Format             |
            +--------+------------+----------------------+----------------+--------------------+
            |        |            |                      |                | Nested (Structure) |
            | 0      | size       | 6                    | 06             | uint8              |
            |        | array      |                      |                | sizedstrarray      |
            |        |   [0]      |                      |                | Simple (Structure) |
            | 1      |     size   | 12                   | 0c             | uint8              |
            | 2      |     string | <insufficient bytes> | 48 65 6c 6c 6f | ascii              |
            +--------+------------+----------------------+----------------+--------------------+

            InsufficientMemoryError occurred during unpack operation:

            7 too few bytes to unpack ascii, 12 needed, only 5 available
            """
        )

        assert wrap_message(trap.value) == expected

    def test_smaller_first_limit_string(self):
        """Test outer limit more restrictive than inner limit (string case)."""

        class NestedStr(Structure):
            """Nested string."""

            size: int = member(fmt=uint8, compute=True)
            simple: Simple = sized_member(size=size, fmt=Simple)
            bookend: int = member(fmt=uint8)

        memory_bytes = b"\x06\x0cHello World!\x99"

        with pytest.raises(UnpackError) as trap:
            unpack(NestedStr, memory_bytes)

        expected = Baseline(
            """
            +--------+----------+----------------------+----------------+-----------------------+
            | Offset | Access   | Value                | Bytes          | Format                |
            +--------+----------+----------------------+----------------+-----------------------+
            |        |          |                      |                | NestedStr (Structure) |
            | 0      | size     | 6                    | 06             | uint8                 |
            |        | simple   |                      |                | Simple (Structure)    |
            | 1      |   size   | 12                   | 0c             | uint8                 |
            | 2      |   string | <insufficient bytes> | 48 65 6c 6c 6f | ascii                 |
            +--------+----------+----------------------+----------------+-----------------------+

            InsufficientMemoryError occurred during unpack operation:

            7 too few bytes to unpack ascii, 12 needed, only 5 available
            """
        )

        assert wrap_message(trap.value) == expected


class TestNoCompute(Case):

    """Test case where size member is not set to compute=True."""

    class Struct(Structure):
        size: int = member(fmt=uint8)
        string: str = sized_member(size=size, fmt=ascii_greedy)
        bookend: int = member(fmt=uint8)

    data = CaseData(
        fmt=Struct,
        bindata=b"\x0cHello World!\x99",
        nbytes=None,
        values=(Struct(size=12, string="Hello World!", bookend=0x99),),
        dump=Baseline(
            """
            +--------+----------+----------------+-------------------------------------+--------------------+
            | Offset | Access   | Value          | Bytes                               | Format             |
            +--------+----------+----------------+-------------------------------------+--------------------+
            |        |          |                |                                     | Struct (Structure) |
            |  0     | size     | 12             | 0c                                  | uint8              |
            |        | string   |                |                                     | ascii              |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
            | 13     | bookend  | 153            | 99                                  | uint8              |
            +--------+----------+----------------+-------------------------------------+--------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+----------+----------------+-------------------------------------+--------------------+
            | Offset | Access   | Value          | Bytes                               | Format             |
            +--------+----------+----------------+-------------------------------------+--------------------+
            |        |          |                |                                     | Struct (Structure) |
            |  0     | size     | 12             | 0c                                  | uint8              |
            |        | string   |                |                                     | ascii              |
            |  1     |   [0:12] | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
            | 13     | bookend  | 153            | 99                                  | uint8              |
            +--------+----------+----------------+-------------------------------------+--------------------+
            | 14     |          | <excess bytes> | 99                                  |                    |
            +--------+----------+----------------+-------------------------------------+--------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
        +--------+----------+----------------------+-------------------------------------+--------------------+
        | Offset | Access   | Value                | Bytes                               | Format             |
        +--------+----------+----------------------+-------------------------------------+--------------------+
        |        |          |                      |                                     | Struct (Structure) |
        |  0     | size     | 12                   | 0c                                  | uint8              |
        |        | string   |                      |                                     | ascii              |
        |  1     |   [0:12] | 'Hello World!'       | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |                    |
        |        | bookend  | <insufficient bytes> |                                     | uint8              |
        +--------+----------+----------------------+-------------------------------------+--------------------+

        InsufficientMemoryError occurred during unpack operation:

        1 too few bytes to unpack uint8, 1 needed, only 0 available
        """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, size: int, string: str, bookend: int) -> None:
                self[:] = (size, string, bookend)

            @size.getter
            def size(self) -> int:
                return self[0]

            @size.setter
            def size(self, value: int) -> None:
                self[0] = value

            @string.getter
            def string(self) -> str:
                return self[1]

            @string.setter
            def string(self, value: str) -> None:
                self[1] = value

            @bookend.getter
            def bookend(self) -> int:
                return self[2]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_size, m_string, m_bookend) = value

                if dump is None:
                    cls.size.fmt.__pack__(m_size, pieces, dump)

                    cls.string.fmt.__pack__(m_string, pieces, dump)

                    cls.bookend.fmt.__pack__(m_bookend, pieces, dump)

                else:
                    size_dump = dump.add_record(access="size", fmt=cls.size.fmt)
                    cls.size.fmt.__pack__(m_size, pieces, size_dump)

                    string_dump = dump.add_record(access="string", fmt=cls.string.fmt)
                    cls.string.fmt.__pack__(m_string, pieces, string_dump)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    cls.bookend.fmt.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Struct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_size, offset = cls.size.fmt.__unpack__(buffer, offset, dump)

                    buffer, string_buffer = buffer[:offset + m_size], buffer
                    if len(buffer) < offset + m_size:
                        cls.string.report_insufficient_bytes(buffer, offset, m_size)
                    m_string, offset = cls.string.fmt.__unpack__(buffer, offset, dump)
                    if offset < len(buffer):
                        cls.string.report_extra_bytes(buffer[offset:])
                    buffer = string_buffer

                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, dump)

                else:
                    size_dump = dump.add_record(access="size", fmt=cls.size.fmt)
                    m_size, offset = cls.size.fmt.__unpack__(buffer, offset, size_dump)

                    string_dump = dump.add_record(access="string", fmt=cls.string.fmt)
                    buffer, string_buffer = buffer[:offset + m_size], buffer
                    if len(buffer) < offset + m_size:
                        cls.string.report_insufficient_bytes(buffer, offset, m_size, string_dump)
                    m_string, offset = cls.string.fmt.__unpack__(buffer, offset, string_dump)
                    if offset < len(buffer):
                        cls.string.report_extra_bytes(buffer[offset:], string_dump)
                    buffer = string_buffer

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, bookend_dump)

                structure[:] = (m_size, m_string, m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(size={self.size!r}, string={self.string!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestComputeNoSetter:

    """Test where size member is computed but sized member is r/o or explict setter."""

    def test_readonly(self):
        if sys.version_info < (3, 11):
            expected_message = Baseline(
                """
                can't set attribute 'string'
                """
            )
        else:
            expected_message = Baseline(
                """
                property 'string' of 'TestComputeNoSetter.test_readonly.<locals>.Struct' object has no setter
                """
            )

        class Struct(Structure):
            size: int = member(fmt=uint8, compute=True)
            string: str = sized_member(size=size, fmt=ascii_greedy, readonly=True)

        struct = Struct(string="hello world")

        with pytest.raises(AttributeError) as trap:
            struct.string = "hello"

        actual_message = str(trap.value)

        if sys.version_info < (3, 10):
            actual_message += " 'string'"

        assert actual_message == expected_message

    def test_explicit_setter(self):
        """Ensure size member not messed with explict setter."""

        class Struct(Structure):
            size: int = member(fmt=uint8, compute=True)
            string: str = sized_member(size=size, fmt=ascii_greedy)

            @string.setter
            def string(self, value):
                self[1] = value

        struct = Struct(size=5, string="hello")
        struct.string = ""

        assert struct.size == 5


class TestComputeGetter:

    """Test where size member is computed and has explict getter."""

    def test_getter(self):
        """Ensure size member getter not messed with."""

        class Struct(Structure):
            size: int = member(fmt=uint8, compute=True)

            @size.getter
            def size(self):
                return 99

            string: str = sized_member(size=size, fmt=ascii_greedy)

        struct = Struct(string="hello")

        assert struct.size == 99



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test get/set item (indexing) BitFields data store transform."""

import pytest
from baseline import Baseline

from plum.bitfields import BitFields
from plum.conformance import wrap_message


class Sample(BitFields, nbytes=1):

    """No bit fields."""


class TestHappyPath:
    def test_get(self):
        sample = Sample.from_int(0xAB)
        assert sample[:] == [True, True, False, True, False, True, False, True]
        assert sample[0] is True
        assert sample[2] is False
        assert sample[7] is True

    def test_set_single_with_bool(self):
        sample = Sample.from_int(0)
        sample[0] = True
        assert int(sample) == 1

    def test_reset_single_with_bool(self):
        sample = Sample.from_int(0xFF)
        sample[0] = False
        assert int(sample) == 0xFE

    def test_set_range(self):
        sample = Sample.from_int(0)
        sample[1:5] = [True, False, True, True]
        assert int(sample) == 0x1A

    def test_reset_range(self):
        sample = Sample.from_int(255)
        sample[1:5] = [False, True, False, False]
        assert int(sample) == 0xE5


class TestExceptions:
    def test_wrong_slice_size(self):
        expected_massage = Baseline(
            """
            slice and value not same length
            """
        )
        sample = Sample.from_int(0)
        with pytest.raises(ValueError) as trap:
            sample[0:1] = [True, True]

        assert wrap_message(trap.value) == expected_massage



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Data view base class."""

from typing import Any, Tuple, Type, Union

from math import ceil, floor

from .data import Data, DataMeta
from .transform import Transform
from .dump import Dump
from .exceptions import ImplementationError, UnpackError


class PlumView:

    """Data view base class.

    :param DataMeta fmt: associated plum type
    :param buffer: bytes buffer
    :type buffer: bytes-like (e.g. bytes, bytearray, memoryview)
    :param int offset: byte offset

    """

    # instance attributes (set as class attributes for pylint)
    __type__: Type[Data] = Data
    __offset__ = 0
    __buffer__ = bytearray()
    __fob__: Tuple[Type[Data], int, bytearray] = (__type__, __offset__, __buffer__)

    def __init__(self, fmt: Union[Type[Data], Transform], buffer, offset):
        if not isinstance(fmt, (DataMeta, Transform)):
            raise TypeError("invalid plumtype")

        object.__setattr__(self, "__type__", fmt)
        object.__setattr__(self, "__buffer__", buffer)
        object.__setattr__(self, "__offset__", offset)
        object.__setattr__(self, "__fob__", (fmt, offset, buffer))

    def __repr__(self):
        try:
            value = self.unpack()
        except Exception:  # pylint: disable=broad-except
            value = f"<view at 0x{self.__offset__:x}>"
        else:
            value = f"<view at 0x{self.__offset__:x}: {value!r}>"
        return value

    def __str__(self):
        try:
            value = str(self.unpack())
        except Exception:  # pylint: disable=broad-except
            value = f"<view at 0x{self.__offset__:x}>"
        return value

    def cast(self, cls: Type[Data]):
        """Create a new view of item's buffer bytes.

        :param DataMeta cls: view type
        :returns: new view
        :rtype: cls
        """
        return cls.view(self.__buffer__, self.__offset__)

    @property
    def dump(self):
        """Packed bytes summary.

        :returns: summary table of view detailing bytes and layout
        :rtype: str

        """
        return self.unpack_and_dump()[1]

    def unpack(self) -> Any:
        """Unpack item from buffer bytes."""
        fmt, offset, buffer = self.__fob__
        try:
            # None -> dump
            return fmt.__unpack__(buffer, offset, None)[0]

        except Exception as exc:
            self.unpack_and_dump()
            raise ImplementationError() from exc  # pragma: no cover

    def unpack_and_dump(self) -> Tuple[Data, Dump]:
        r"""Unpack item from bytes and produce packed bytes summary.

        :raises: ``UnpackError`` if insufficient bytes, excess bytes, or value error

        """
        fmt, offset, buffer = self.__fob__

        dump = Dump(offset=self.__offset__)
        try:
            item, _offset = fmt.__unpack__(
                buffer, offset, dump.add_record(fmt=fmt.name)
            )

        except Exception as exc:
            raise UnpackError(dump, exc) from exc

        return item, dump

    @property
    def nbytes(self) -> int:
        """Bytes buffer view size in bytes."""
        nbytes = self.__type__.__nbytes__

        # FUTURE - remove cover comment when views support variable sized types
        if nbytes is None:  # pragma: no cover
            fmt, offset, buffer = self.__fob__
            try:
                # None -> dump
                _item, end = fmt.__unpack__(buffer, offset, None)

            except Exception as exc:
                self.unpack_and_dump()
                raise ImplementationError() from exc  # pragma: no cover

            nbytes = end - offset

        return nbytes

    def pack(self) -> bytes:
        """Pack as formatted bytes.

        :raises: ``PackError`` if type error, value error, etc.

        """
        fmt, offset, buffer = self.__fob__

        nbytes = fmt.__nbytes__

        # FUTURE - remove cover comment when views support variable sized types
        if nbytes is None:  # pragma: no cover
            try:
                # None -> dump
                _item, end = fmt.__unpack__(buffer, offset, None)

            except Exception as exc:
                self.unpack_and_dump()
                raise ImplementationError() from exc  # pragma: no cover

            bindata = bytes(buffer[offset:end])
        else:
            bindata = bytes(buffer[offset : offset + nbytes])

        if len(bindata) != nbytes:
            self.pack_and_dump()
            raise ImplementationError()  # pragma: no cover

        return bindata

    ipack = pack

    def pack_and_dump(self):
        """Pack value as formatted bytes and produce bytes summary.

        :returns: bytes buffer, packed bytes summary
        :rtype: bytearray, Dump

        """
        fmt, offset, buffer = self.__fob__

        dump = Dump(offset=self.__offset__)
        try:
            _item, end = fmt.__unpack__(buffer, offset, dump.add_record(fmt=fmt))

        except Exception as exc:
            raise UnpackError(dump, exc) from exc

        return bytes(buffer[offset:end]), dump

    ipack_and_dump = pack_and_dump

    def set(self, value):
        """Pack value into bytes buffer.

        :param object value: new value

        """
        fob, offset, buffer = self.__fob__

        membytes = fob.pack(value)

        buffer[offset : offset + len(membytes)] = membytes


class NumberView(PlumView):

    """Numeric view class."""

    def __abs__(self):
        return abs(self.unpack())

    def __add__(self, other):
        return self.unpack() + other

    def __divmod__(self, other):
        return divmod(self.unpack(), other)

    def __float__(self):
        return float(self.unpack())

    def __eq__(self, other):
        return self.unpack() == other

    def __ge__(self, other):
        return self.unpack() >= other

    def __gt__(self, other):
        return self.unpack() > other

    def __iadd__(self, other):
        self.set(self.unpack() + other)
        return self

    def __imod__(self, other):
        self.set(self.unpack() % other)
        return self

    def __imul__(self, other):
        self.set(self.unpack() * other)
        return self

    def __int__(self):
        return int(self.unpack())

    def __ipow__(self, other, modulus=None):
        self.set(self.unpack().__pow__(other, modulus))  # pylint: disable=no-member
        return self

    def __isub__(self, other):
        self.set(self.unpack() - other)
        return self

    def __le__(self, other):
        return self.unpack() <= other

    def __lt__(self, other):
        return self.unpack() < other

    def __mod__(self, other):
        return self.unpack() % other

    def __mul__(self, other):
        return self.unpack() * other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __neg__(self):
        return -self.unpack()  # pylint: disable=invalid-unary-operand-type

    def __pos__(self):
        return +self.unpack()  # pylint: disable=invalid-unary-operand-type

    def __pow__(self, exponent, modulus=None):
        return self.unpack().__pow__(exponent, modulus)  # pylint: disable=no-member

    def __radd__(self, other):
        return other + self.unpack()

    def __rdivmod__(self, other):
        return divmod(other, self.unpack())

    def __repr__(self):
        try:
            value = self.unpack()
        except Exception:  # pylint: disable=broad-except
            rep = f"<view at 0x{self.__offset__:x}>"
        else:
            rep = f"<{self.__type__.name} view at 0x{self.__offset__:x}: {value}>"
        return rep

    def __rmod__(self, other):
        return other % self.unpack()

    def __rmul__(self, other):
        return other * self.unpack()

    def __rpow__(self, other):
        return other ** self.unpack()

    def __rsub__(self, other):
        return other - self.unpack()

    def __rtruediv__(self, other):
        return other / self.unpack()

    def __sub__(self, other):
        return self.unpack() - other

    def __truediv__(self, other):
        return self.unpack() / other

    def __ceil__(self):
        return ceil(self.unpack())

    def __floor__(self):
        return floor(self.unpack())

    def __floordiv__(self, other):
        return self.unpack().__floordiv__(other)  # pylint: disable=no-member

    def __rfloordiv__(self, other):
        return self.unpack().__rfloordiv__(other)  # pylint: disable=no-member

    def __round__(self, ndigits=None):
        return self.unpack().__round__(ndigits)  # pylint: disable=no-member

    def __trunc__(self):
        return self.unpack().__trunc__()  # pylint: disable=no-member

    def __hash__(self):
        raise TypeError(f"unhashable type: {type(self).__name__}")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test pointer views."""

from plum.int import IntX
from plum.littleendian import uint16


pointeruint16 = IntX(name="pointeruint16", nbytes=1, dref=uint16)
pointeruint16pointer = IntX(name="pointeruint16pointer", nbytes=1, dref=pointeruint16)


class TestBasics:

    """Test basic features of pointer views."""

    def test_int_behavior(self):
        """Test that it behaves as an integer."""
        buffer = bytearray([0, 4, 0, 0, 1, 2, 3, 4])
        pointer = pointeruint16.view(buffer, 1)
        assert pointer == 4

        pointer.set(2)
        assert buffer == bytearray([0, 2, 0, 0, 1, 2, 3, 4])

    def test_dref_read(self):
        """Test dereferencing produces correct value."""
        buffer = bytearray([0, 4, 0, 0, 1, 2, 3, 4])
        pointer = pointeruint16.view(buffer, 1)
        assert pointer[0] == 0x201
        assert pointer[1] == 0x403

        # change data behind the scenes (last byte)
        buffer[7] = 0x99
        assert pointer[1] == 0x9903

        # change pointer behind the scenes
        buffer[1] = 3
        assert pointer[0] == 0x100
        assert pointer[1] == 0x302

    def test_dref_write(self):
        """Test dereferencing with assignment writes bytes correctly."""
        buffer = bytearray([0, 4, 0, 0, 0, 0, 0, 0])
        pointer = pointeruint16.view(buffer, 1)

        pointer[0] = 0x201
        assert buffer == bytearray([0, 4, 0, 0, 1, 2, 0, 0])

        pointer[1] = 0x403
        assert buffer == bytearray([0, 4, 0, 0, 1, 2, 3, 4])

    def test_nesting(self):
        """Test nested pointers."""
        buffer = bytearray([0, 4, 0, 0, 6, 8, 1, 2, 3, 4])
        pointer = pointeruint16pointer.view(buffer, 1)

        assert pointer == 4
        assert pointer[0] == 6
        assert pointer[0][0] == 0x201
        assert pointer[0][1] == 0x403
        assert pointer[1][0] == 0x403



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test miscellaneous StrX transform features."""

import pytest

from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import PackError, UnpackError
from plum.str import StrX
from plum.utilities import pack, pack_and_dump, unpack, unpack_and_dump

ascii_greedy = StrX(name="ascii_greedy", encoding="ascii")
ascii_zero = StrX(name="ascii_zero", encoding="ascii", zero_termination=True)
custom = StrX(name="custom", encoding="ascii")
utf8 = StrX(name="utf8", encoding="utf8")
ascii_ignore = StrX(name="ascii_ignore", encoding="ascii", errors="ignore")


class TestUnpack:

    """Test various unpack() corner cases."""

    def test_zero_term_decoding_error(self):
        """Test unpacking a zero terminated string that has an invalid encoding."""

        with pytest.raises(UnpackError) as trap:
            unpack(ascii_zero, b"hello\x80world\x00")

        expected = Baseline(
            """
            +--------+-----------+---------+----------------------+------------+
            | Offset | Access    | Value   | Bytes                | Format     |
            +--------+-----------+---------+----------------------+------------+
            |        |           |         |                      | ascii_zero |
            |  0     | [0:5]     | 'hello' | 68 65 6c 6c 6f       |            |
            |  5     | --error-- |         | 80 77 6f 72 6c 64 00 |            |
            +--------+-----------+---------+----------------------+------------+

            UnicodeDecodeError occurred during unpack operation:

            'ascii' codec can't decode byte 0x80 in position 0: ordinal not in
            range(128)
            """
        )

        assert wrap_message(trap.value) == expected

    def test_multibyte_chars(self):
        """Test unpacking characters that are more than one byte."""
        sample, dump = unpack_and_dump(
            utf8, b"Mohu j\xc3\xadst sklo, neubl\xc3\xad\xc5\xbe\xc3\xad mi."
        )
        assert sample == "Mohu jíst sklo, neublíží mi."

        # pylint: disable=line-too-long
        expected = Baseline(
            """
            +--------+---------+-------------------+-------------------------------------------------+--------+
            | Offset | Access  | Value             | Bytes                                           | Format |
            +--------+---------+-------------------+-------------------------------------------------+--------+
            |        |         |                   |                                                 | utf8   |
            |  0     | [0:15]  | 'Mohu j\xedst sklo,' | 4d 6f 68 75 20 6a c3 ad 73 74 20 73 6b 6c 6f 2c |        |
            | 16     | [15:28] | ' neubl\xed\u017e\xed mi.'   | 20 6e 65 75 62 6c c3 ad c5 be c3 ad 20 6d 69 2e |        |
            +--------+---------+-------------------+-------------------------------------------------+--------+
            """
        )

        assert str(dump) == expected

    def test_multibyte_chars_byte_shortage(self):
        """Test unpacking characters that are more than one byte but short bytes."""
        with pytest.raises(UnpackError) as trap:
            unpack_and_dump(utf8, b"Mohu j\xc3")

        # pylint: disable=line-too-long
        expected = Baseline(
            """
            +--------+--------+----------+----------------------+--------+
            | Offset | Access | Value    | Bytes                | Format |
            +--------+--------+----------+----------------------+--------+
            |        |        |          |                      | utf8   |
            | 0      | [0:6]  | 'Mohu j' | 4d 6f 68 75 20 6a c3 |        |
            +--------+--------+----------+----------------------+--------+

            UnicodeDecodeError occurred during unpack operation:

            'utf-8' codec can't decode byte 0xc3 in position 0: unexpected end of
            data
            """
        )

        assert wrap_message(trap.value) == expected


class TestPack:

    """Test various pack() corner cases."""

    def test_encoding_error(self):
        """Test packing a string that can't be encoded."""
        expected = Baseline(
            r"""
            +--------+-----------+-----------------------+----------------------------+--------------+
            | Offset | Access    | Value                 | Bytes                      | Format       |
            +--------+-----------+-----------------------+----------------------------+--------------+
            |        |           |                       |                            | ascii_greedy |
            | 0      | [0:9]     | 'The quick'           | 54 68 65 20 71 75 69 63 6b |              |
            |        | --error-- | '\x80brown fox jumps' |                            |              |
            |        |           | ' over the lazy d'    |                            |              |
            |        |           | 'og'                  |                            |              |
            +--------+-----------+-----------------------+----------------------------+--------------+

            UnicodeEncodeError occurred during pack operation:

            'ascii' codec can't encode character '\x80' in position 0: ordinal not
            in range(128)
            """
        )

        with pytest.raises(PackError) as trap:
            # include more than 16 characters after the problem to
            # exercise the dump logic to break up the string
            pack("The quick\x80brown fox jumps over the lazy dog", ascii_greedy)

        assert wrap_message(trap.value) == expected

    def test_exceeds_size(self):
        """Test packing a string that exceeds the size limit."""

        sized_ascii = StrX(name="sized_ascii", nbytes=5, encoding="ascii")

        expected = Baseline(
            """
            +--------+--------+----------+-------------------+-------------+
            | Offset | Access | Value    | Bytes             | Format      |
            +--------+--------+----------+-------------------+-------------+
            |        |        |          |                   | sized_ascii |
            | 0      | [0:6]  | 'hello!' | 68 65 6c 6c 6f 21 |             |
            +--------+--------+----------+-------------------+-------------+

            TypeError occurred during pack operation:

            number of string bytes (6) exceeds limit (5) for <transform
            'sized_ascii'>
            """
        )

        with pytest.raises(PackError) as trap:
            pack("hello!", sized_ascii)

        assert wrap_message(trap.value) == expected


class TestPackAndDump:

    """Test various pack_and_dump() corner cases."""

    def test_empty(self):
        """Test empty string pack."""
        expected_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            |        |        |       |       | custom |
            |        | [0:0]  | ''    |       |        |
            +--------+--------+-------+-------+--------+
            """
        )

        buffer, dump = pack_and_dump("", custom)

        assert buffer == b""
        assert str(dump) == expected_dump

    def test_one_full_dump_row(self):
        """Test string exactly size of one dump row."""
        expected_dump = Baseline(
            """
            +--------+--------+--------------------+-------------------------------------------------+--------+
            | Offset | Access | Value              | Bytes                                           | Format |
            +--------+--------+--------------------+-------------------------------------------------+--------+
            |        |        |                    |                                                 | custom |
            |  0     | [0:16] | '0123456789abcdef' | 30 31 32 33 34 35 36 37 38 39 61 62 63 64 65 66 |        |
            +--------+--------+--------------------+-------------------------------------------------+--------+
            """
        )

        sample = "0123456789abcdef"
        buffer, dump = pack_and_dump(sample, custom)

        assert buffer == bytearray(sample, encoding="ascii")
        assert str(dump) == expected_dump

    def test_two_full_dump_row(self):
        """Test string exactly size of one dump row."""
        expected_dump = Baseline(
            """
            +--------+---------+--------------------+-------------------------------------------------+--------+
            | Offset | Access  | Value              | Bytes                                           | Format |
            +--------+---------+--------------------+-------------------------------------------------+--------+
            |        |         |                    |                                                 | custom |
            |  0     | [0:16]  | '0123456789abcdef' | 30 31 32 33 34 35 36 37 38 39 61 62 63 64 65 66 |        |
            | 16     | [16:32] | '0123456789abcdef' | 30 31 32 33 34 35 36 37 38 39 61 62 63 64 65 66 |        |
            +--------+---------+--------------------+-------------------------------------------------+--------+
            """
        )

        sample = "0123456789abcdef" * 2
        buffer, dump = pack_and_dump(sample, custom)

        assert buffer == bytearray(sample, encoding="ascii")
        assert str(dump) == expected_dump


class TestErrors:

    """Test "errors" property."""

    def test_replace(self):
        assert ascii_ignore.unpack(b"hello\x80world") == "helloworld"


class TestFixedSizeZeroTerm:

    ascii_zero = StrX(
        name="ascii_zero",
        encoding="ascii",
        zero_termination=True,
        nbytes=10,
        pad=b"\x00",
    )

    def test_unpack(self):
        assert self.ascii_zero.unpack(b"hello\x00fill") == "hello"

    def test_unpack_and_dump(self):
        expected_dump = Baseline(
            """
            +--------+-----------------+---------+----------------+------------+
            | Offset | Access          | Value   | Bytes          | Format     |
            +--------+-----------------+---------+----------------+------------+
            |        |                 |         |                | ascii_zero |
            |  0     | [0:5]           | 'hello' | 68 65 6c 6c 6f |            |
            |  5     | --termination-- |         | 00             |            |
            |  6     | --pad--         |         | 66 69 6c 6c    |            |
            +--------+-----------------+---------+----------------+------------+
            """
        )

        string, dump = self.ascii_zero.unpack_and_dump(b"hello\x00fill")

        assert str(dump) == expected_dump
        assert string == "hello"

    def test_pack(self):
        assert self.ascii_zero.pack("hello") == b"hello" + bytes(5)

    def test_pack_and_dump(self):
        expected_dump = Baseline(
            """
            +--------+-----------------+---------+----------------+------------+
            | Offset | Access          | Value   | Bytes          | Format     |
            +--------+-----------------+---------+----------------+------------+
            |        |                 |         |                | ascii_zero |
            |  0     | [0:5]           | 'hello' | 68 65 6c 6c 6f |            |
            |  5     | --termination-- |         | 00             |            |
            |  6     | --pad--         |         | 00 00 00 00    |            |
            +--------+-----------------+---------+----------------+------------+
            """
        )

        buffer, dump = self.ascii_zero.pack_and_dump("hello")

        assert str(dump) == expected_dump
        assert buffer == b"hello" + bytes(5)


class TestFixedSizeZeroTermNoPad:

    ascii_zero = StrX(
        name="ascii_zero", encoding="ascii", zero_termination=True, nbytes=10
    )

    unpack_message = Baseline(
        """


        +--------+----------------------+-------------------+------------+
        | Offset | Value                | Bytes             | Format     |
        +--------+----------------------+-------------------+------------+
        | 0      | <insufficient bytes> | 68 65 6c 6c 6f 00 | ascii_zero |
        +--------+----------------------+-------------------+------------+

        InsufficientMemoryError occurred during unpack operation:

        4 too few bytes to unpack ascii_zero, 10 needed, only 6 available
        """
    )

    def test_unpack(self):
        with pytest.raises(UnpackError) as trap:
            self.ascii_zero.unpack(b"hello\x00")

        assert str(trap.value) == self.unpack_message

    def test_unpack_and_dump(self):
        with pytest.raises(UnpackError) as trap:
            self.ascii_zero.unpack_and_dump(b"hello\x00")

        assert str(trap.value) == self.unpack_message

    pack_message = Baseline(
        """


        +--------+-----------------+---------+----------------+------------+
        | Offset | Access          | Value   | Bytes          | Format     |
        +--------+-----------------+---------+----------------+------------+
        |        |                 |         |                | ascii_zero |
        | 0      | [0:5]           | 'hello' | 68 65 6c 6c 6f |            |
        | 5      | --termination-- |         | 00             |            |
        +--------+-----------------+---------+----------------+------------+

        ValueError occurred during pack operation:

        number of string bytes (6) falls short of (10) for <transform 'ascii_zero'>
        """
    )

    def test_pack(self):
        with pytest.raises(PackError) as trap:
            self.ascii_zero.pack("hello")

        assert str(trap.value) == self.pack_message

    def pack_and_dump(self):
        with pytest.raises(PackError) as trap:
            self.ascii_zero.pack_and_dump("hello")

        assert str(trap.value) == self.pack_message


class TestPadNotNeeded:

    ascii_fixed = StrX(name="ascii_fixed", encoding="ascii", nbytes=5, pad=b"\x00")

    def test_pack(self):
        assert self.ascii_fixed.pack("hello") == b"hello"

    def test_pack_and_dump(self):
        expected_dump = Baseline(
            """
            +--------+--------+---------+----------------+-------------+
            | Offset | Access | Value   | Bytes          | Format      |
            +--------+--------+---------+----------------+-------------+
            |        |        |         |                | ascii_fixed |
            | 0      | [0:5]  | 'hello' | 68 65 6c 6c 6f |             |
            +--------+--------+---------+----------------+-------------+
            """
        )

        buffer, dump = self.ascii_fixed.pack_and_dump("hello")

        assert str(dump) == expected_dump
        assert buffer == b"hello"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Object to bytes and bytes to Object transform."""

from typing import Any, Callable, List, Optional, Tuple

from ._getbytes import getbytes
from .dump import Record
from .exceptions import InsufficientMemoryError
from .transform import Transform


PackType = Callable[[Any], bytes]
UnpackType = Callable[[bytes], Any]


class ObjectX(Transform):

    """Object to bytes and bytes to object transform."""

    __nbytes__: int
    __npu__: Tuple[int, PackType, UnpackType]

    def __init__(
        self,
        pack: PackType,
        unpack: UnpackType,
        nbytes: int,
        name: str = "object",
        hint: str = "Any",
    ) -> None:
        super().__init__(name, hint)

        assert nbytes >= 0

        self.__nbytes__ = nbytes
        self.__npu__ = nbytes, pack, unpack

    def __pack__(
        self, value: int, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        _nbytes, pack, _unpack = self.__npu__

        if dump is None:
            pieces.append(pack(value))
        else:
            dump.value = repr(value)

            piece = dump.memory = pack(value)

            pieces.append(piece)

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[int, int]:
        nbytes, _pack, unpack = self.__npu__

        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError("too few bytes to unpack")

            return unpack(buffer[offset:end]), end

        chunk, end = getbytes(buffer, offset, dump, nbytes)

        value = unpack(chunk)

        dump.value = repr(value)

        return value, end



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Transform conformance test case base class."""

from textwrap import wrap
from typing import Any, Generator, List, Optional, Tuple, Type, Union

from .bigendian import uint8
from .buffer import Buffer
from .data import Data
from .exceptions import (
    ExcessMemoryError,
    InsufficientMemoryError,
    SizeError,
    UnpackError,
)
from .transform import Transform
from .utilities import (
    pack,
    pack_and_dump,
    unpack,
    unpack_and_dump,
)

# pylint: disable=unidiomatic-typecheck,no-member


class CaseData:  # pylint: disable=too-many-instance-attributes

    """Conformance test case stimulus and expectations."""

    fmt: Union[Type[Data], Transform]
    """Transform to test."""

    bindata: bytes
    """Expected packed bytes."""

    values: Tuple[Any, ...]
    """Expected value when unpacked, followed by alternative packable values."""

    dump: str
    """Expected dump of sample plum type instance."""

    nbytes: Optional[int]
    """Size of plum type being tested."""

    excess: str
    """Expected exception message of unpack() method when 1 byte extra."""

    shortage: str
    """Expected exception message of unpack() method when short 1 byte."""

    implementation: Optional[str] = None
    """Expected generated implementation code."""

    def __init__(
        self,
        *,
        fmt: Union[Type[Data], Transform],
        bindata: bytes,
        values: Tuple[Any, ...],
        dump: str,
        nbytes: Optional[int],
        excess: str,
        shortage: str,
        implementation: Optional[str] = None,
    ):
        self.fmt = fmt
        self.bindata = bindata
        self.values = values
        self.dump = dump
        self.nbytes = nbytes
        self.excess = excess
        self.shortage = shortage
        self.implementation = implementation


class Case:  # pylint: disable=too-many-public-methods

    """Test basic API conformance and utility usage."""

    data: CaseData

    def test_transform_nbytes_property(self):
        """Test transform's nbytes property."""
        try:
            nbytes = self.data.fmt.nbytes
        except SizeError:
            nbytes = None

        assert nbytes == self.data.nbytes

    def test_data_store_nbytes_property(self):
        """Test data store nbytes property."""
        if not isinstance(self.data.fmt, Transform) and issubclass(self.data.fmt, Data):
            assert unpack(self.data.fmt, self.data.bindata).nbytes == len(
                self.data.bindata
            )

    def test_data_store_dump(self):
        """Test data store dump property."""
        if not isinstance(self.data.fmt, Transform) and issubclass(self.data.fmt, Data):
            assert str(unpack(self.data.fmt, self.data.bindata).dump) == self.data.dump

    def test_utility_pack(self):
        """Test utility pack() function."""
        for value in self.data.values:
            assert pack(value, self.data.fmt) == self.data.bindata

    def test_transform_pack(self):
        """Test transform pack() method."""
        for value in self.data.values:
            assert self.data.fmt.pack(value) == self.data.bindata

    def test_data_store_ipack(self):
        """Test data store ipack() method."""
        if not isinstance(self.data.fmt, Transform) and issubclass(self.data.fmt, Data):
            assert unpack(self.data.fmt, self.data.bindata).ipack() == self.data.bindata

    def test_utility_pack_and_dump(self):
        """Test utility pack_and_dump() function."""
        for value in self.data.values:
            buffer, dump = pack_and_dump(value, self.data.fmt)
            assert str(dump) == self.data.dump
            assert buffer == self.data.bindata

    def test_transform_pack_and_dump(self):
        """Test transfrom pack_and_dump() method."""
        for value in self.data.values:
            buffer, dump = self.data.fmt.pack_and_dump(value)
            assert str(dump) == self.data.dump
            assert buffer == self.data.bindata

    def test_data_store_ipack_and_dump(self):
        """Test data store ipack_and_dump() method."""
        if not isinstance(self.data.fmt, Transform) and issubclass(self.data.fmt, Data):
            datastore = unpack(self.data.fmt, self.data.bindata)
            buffer, dump = datastore.ipack_and_dump()
            assert str(dump) == self.data.dump
            assert buffer == self.data.bindata

    def test_utility_unpack(self):
        """Test utility unpack() function."""
        expected_value = self.data.values[0]
        item = unpack(self.data.fmt, self.data.bindata)
        assert type(item) is type(expected_value)
        assert item == expected_value

    def test_transform_unpack(self):
        """Test transform unpack() method."""
        expected_value = self.data.values[0]
        item = self.data.fmt.unpack(self.data.bindata)
        assert type(item) is type(expected_value)
        assert item == expected_value

    def test_buffer_unpack(self):
        """Test Buffer.unpack() method (can unpack from the middle)."""
        try:
            self.data.fmt.nbytes
        except SizeError:
            # must be greedy transform
            with Buffer(b"\x00" + self.data.bindata) as buffer:
                head = buffer.unpack(uint8)
                item = buffer.unpack(self.data.fmt)
                tail = 0x99
        else:
            with Buffer(b"\x00" + self.data.bindata + b"\x99") as buffer:
                head = buffer.unpack(uint8)
                item = buffer.unpack(self.data.fmt)
                tail = buffer.unpack(uint8)

        expected_value = self.data.values[0]

        assert head == 0
        assert type(item) is type(expected_value)
        assert item == expected_value
        assert tail == 0x99

    def test_utility_unpack_and_dump(self):
        """Test utility unpack_and_dump() function."""
        expected_value = self.data.values[0]
        item, dump = unpack_and_dump(self.data.fmt, self.data.bindata)
        assert str(dump) == self.data.dump
        assert type(item) is type(expected_value)
        assert item == expected_value

    def test_transform_unpack_and_dump(self):
        """Test transform unpack_and_dump() method."""
        expected_value = self.data.values[0]
        item, dump = self.data.fmt.unpack_and_dump(self.data.bindata)
        assert str(dump) == self.data.dump
        assert type(item) is type(expected_value)
        assert item == expected_value

    def test_buffer_unpack_and_dump(self):
        """Test Buffer.unpack_and_dump() method (can unpack from the middle)."""
        try:
            self.data.fmt.nbytes
        except SizeError:
            # must be greedy transform
            with Buffer(b"\x00" + self.data.bindata) as buffer:
                head = buffer.unpack(uint8)
                item, dump = buffer.unpack_and_dump(self.data.fmt)
                tail = 0x99
        else:
            with Buffer(b"\x00" + self.data.bindata + b"\x99") as buffer:
                head = buffer.unpack(uint8)
                item, dump = buffer.unpack_and_dump(self.data.fmt)
                tail = buffer.unpack(uint8)

        expected_value = self.data.values[0]

        # change dump offset to make consistent with baselined dump
        assert dump.offset == 1
        dump.offset = 0

        assert str(dump) == self.data.dump
        assert head == 0
        assert type(item) is type(expected_value)
        assert item == expected_value
        assert tail == 0x99

    def test_unpack_shortage(self):
        """Test unpack() usage with insufficient bytes."""
        if self.data.shortage not in {"N/A"} and self.data.bindata:
            try:
                unpack(self.data.fmt, self.data.bindata[:-1])
            except UnpackError as exception:
                assert wrap_message(exception) == self.data.shortage
                assert isinstance(exception.__context__, InsufficientMemoryError)
            else:  # pragma: no cover
                raise AssertionError("unpack() did not raise an exception")

            if not isinstance(self.data.fmt, Transform) and issubclass(
                self.data.fmt, Data
            ):
                try:
                    self.data.fmt.unpack(self.data.bindata[:-1])
                except UnpackError as exception:
                    assert wrap_message(exception) == self.data.shortage
                    assert isinstance(exception.__context__, InsufficientMemoryError)
                else:  # pragma: no cover
                    raise AssertionError("unpack() did not raise an exception")

    def test_unpack_excess(self):
        """Test unpack() usage with excess bytes."""
        if self.data.excess not in {"N/A"}:
            try:
                unpack(self.data.fmt, self.data.bindata + b"\x99")
            except UnpackError as exception:
                assert wrap_message(exception) == self.data.excess
                assert isinstance(exception.__context__, ExcessMemoryError)
            else:  # pragma: no cover
                raise AssertionError("unpack() did not raise an exception")

            if not isinstance(self.data.fmt, Transform) and issubclass(
                self.data.fmt, Data
            ):
                try:
                    self.data.fmt.unpack(self.data.bindata + b"\x99")
                except UnpackError as exception:
                    assert wrap_message(exception) == self.data.excess
                    assert isinstance(exception.__context__, ExcessMemoryError)
                else:  # pragma: no cover
                    raise AssertionError("unpack() did not raise an exception")

    def test_implementation(self):
        """Test generated code matches baseline."""
        if self.data.implementation is not None:
            assert self.data.fmt.implementation == self.data.implementation


def wrap_message(exc: BaseException) -> str:
    """Line wrap exception message.

    Wrap exception message lines that exceed 80 characters.
    Detect and leave undisturbed `dump()` tables.

    """
    lines_out = []
    queue: List[str] = []
    for line in str(exc).split("\n"):
        stripped_line = line.strip()
        if not stripped_line or (stripped_line[0] in "+|"):
            lines_out.extend(wrap("\n".join(queue)))
            lines_out.append(line)
            queue = []
        else:
            queue.append(line)

    if queue:  # pragma: no cover
        lines_out.extend(wrap("\n".join(queue)))

    while lines_out and not lines_out[0]:
        lines_out.pop(0)

    return "\n".join(lines_out)


def _iter_lines(code: str, methodname: str) -> Generator[str, None, None]:
    lines = code.split("\n")

    active = False

    pattern = f"def {methodname}"

    for line in lines:
        lstripped_line = line.lstrip()

        if lstripped_line.startswith(pattern):
            active = True
        elif lstripped_line and lstripped_line == line:
            active = False

        if active:
            yield line


def extract_method_code(code: str, methodname: str) -> str:
    """Get code for single method from that of many methods."""
    return "\n".join(_iter_lines(code, methodname))



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test items transform with fmt=None and with tuple pairing."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import PackError
from plum.items import items
from plum.littleendian import uint8
from plum.utilities import pack_and_dump


class Test:

    sample_pair = (1, uint8)

    def test_pack(self):
        assert items.pack(self.sample_pair) == bytes.fromhex("01")

    def test_pack_and_dump_method(self):
        expected_dump = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            | 0      | 1     | 01    | uint8  |
            +--------+-------+-------+--------+
            """
        )

        memory, dump = items.pack_and_dump(self.sample_pair)

        assert memory == bytes.fromhex("01")
        assert str(dump) == expected_dump

    def test_pack_and_dump_utility(self):
        expected_dump = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            | 0      | 1     | 01    | uint8  |
            +--------+-------+-------+--------+
            """
        )

        memory, dump = pack_and_dump(self.sample_pair)

        assert memory == bytes.fromhex("01")
        assert str(dump) == expected_dump

    def test_bad_item_in_method(self):
        expected = Baseline(
            """
            +--------+-------+-------+---------------+
            | Offset | Value | Bytes | Format        |
            +--------+-------+-------+---------------+
            |        | (0,)  |       | items:unknown |
            +--------+-------+-------+---------------+

            ValueError occurred during pack operation:

            expected (value, fmt) pair, but got tuple of length 1
            """
        )

        with pytest.raises(PackError) as trap:
            items.pack((0,))

        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, ValueError)

    def test_bad_item_in_utility(self):
        expected = Baseline(
            """
            +--------+-------+-------+---------+
            | Offset | Value | Bytes | Format  |
            +--------+-------+-------+---------+
            |        | (0,)  |       | unknown |
            +--------+-------+-------+---------+

            ValueError occurred during pack operation:

            expected (value, fmt) pair, but got tuple of length 1
            """
        )

        with pytest.raises(PackError) as trap:
            pack_and_dump((0,))

        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, ValueError)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Transform dict/iterable values to bytes and vice versa."""

import sys
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Type, Tuple, Union

from .bytes import BytesX
from .data import Data, DataMeta
from .dump import Record
from .transform import Transform

bytes_x = BytesX(name="bytes")


class Default(Enum):

    """Default sentinel."""

    UNSPECIFIED = "UNSPECIFIED"


if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import Literal

    DefaultUnspecified = Literal[Default.UNSPECIFIED]

else:  # pragma: no cover
    DefaultUnspecified = Default

ItemsFormat = Union[
    Dict[str, Any], List[Any], Transform, Tuple[Any, ...], Type[Data], None
]
DefaultFormat = Union[
    Dict[str, Any],
    List[Any],
    Transform,
    Tuple[Any, ...],
    Type[Data],
    None,
    DefaultUnspecified,
]


def calcsize(fmt: ItemsFormat) -> Union[int, None]:
    """Calculate format size (return None if not a fixed size)."""
    if fmt is None:
        return None

    if isinstance(fmt, (Transform, DataMeta)):
        return fmt.__nbytes__

    if isinstance(fmt, dict):
        sizes = [calcsize(f) for f in fmt.values()]
    elif isinstance(fmt, (tuple, list)):
        sizes = [calcsize(f) for f in fmt]
    else:
        raise TypeError(f"invalid format {fmt!r}")

    try:
        return sum(sizes)  # type: ignore
    except TypeError:
        # must be None in list indicating something variable size
        return None


class ItemsX(Transform):

    """Value to bytes and bytes to value transformer."""

    _fmt: ItemsFormat

    def __init__(
        self,
        fmt: ItemsFormat = None,
        name: Optional[str] = None,
        hint: Optional[str] = None,
    ) -> None:
        hint = self._get_hint(fmt)

        if name is None:
            name = hint

        super().__init__(name, hint)

        self.__nbytes__ = calcsize(fmt)
        self._fmt = fmt

    def _get_hints(self, formats: Iterable[ItemsFormat]) -> List[str]:
        hints = []
        for fmt in formats:
            if fmt is None:
                hints.append("Any")
            elif isinstance(fmt, (dict, list, tuple)):
                hints.append(self._get_hint(fmt))
            else:
                hints.append(fmt.__hint__)
        return hints

    def _get_hint(self, fmt: ItemsFormat) -> str:
        # pylint: disable=too-many-return-statements
        if fmt is None:
            return "Any"

        if isinstance(fmt, dict):
            hints = set(self._get_hints(fmt.values()))
            if len(hints) == 1:
                return f"Dict[str, {hints.pop()}]"
            return "Dict[str, Any]"

        if isinstance(fmt, tuple):
            return f"Tuple[{', '.join(self._get_hints(fmt))}]"

        if isinstance(fmt, list):
            hints = set(self._get_hints(fmt))
            if len(hints) == 1:
                return f"List[{hints.pop()}]"

            return "List[Any]"

        return fmt.__hint__

    @property
    def fmt(self) -> ItemsFormat:
        """Items format."""
        return self._fmt

    def __pack__(
        self,
        value: Any,
        pieces: List[bytes],
        dump: Optional[Record] = None,
        fmt: DefaultFormat = Default.UNSPECIFIED,
    ) -> None:
        # pylint: disable=too-many-branches,arguments-differ
        if dump is not None:
            self.__pack_and_dump__(value, pieces, dump, fmt)

        else:
            if fmt is Default.UNSPECIFIED:
                fmt = self._fmt

            if fmt is None:
                if isinstance(value, (bytes, bytearray)):
                    pieces.append(value)

                elif isinstance(value, tuple):
                    item, bound_fmt = value
                    self.__pack__(item, pieces, dump, bound_fmt)

                elif isinstance(value, Data):
                    value.__pack__(value, pieces, dump)

                elif isinstance(value, list):
                    for item in value:
                        self.__pack__(item, pieces, dump)

                elif isinstance(value, dict):
                    for item in value.values():
                        self.__pack__(item, pieces, dump)

                else:
                    raise TypeError("invalid value")

            elif isinstance(fmt, dict):
                if not isinstance(value, dict) or (set(value) - set(fmt)):
                    raise TypeError("invalid value")

                for name, subfmt in fmt.items():
                    if isinstance(subfmt, (DataMeta, Transform)):
                        subfmt.__pack__(value[name], pieces, dump)
                    else:
                        # must be dict, list, tuple
                        self.__pack__(value[name], pieces, dump, subfmt)

            elif isinstance(fmt, (tuple, list)):
                if not isinstance(value, (tuple, list)) or len(value) != len(fmt):
                    raise TypeError("invalid value")

                for subfmt, item in zip(fmt, value):
                    if isinstance(subfmt, (DataMeta, Transform)):
                        subfmt.__pack__(item, pieces, dump)
                    else:
                        # must be dict, list, tuple
                        self.__pack__(item, pieces, dump, subfmt)

            else:
                fmt.__pack__(value, pieces, dump)

    def __pack_and_dump__(
        self,
        value: Any,
        pieces: List[bytes],
        dump: Record,
        fmt: DefaultFormat = Default.UNSPECIFIED,
    ):
        # pylint: disable=too-many-statements,too-many-branches
        # pylint: disable=too-many-locals,arguments-differ
        if fmt is Default.UNSPECIFIED:
            fmt = self._fmt
            dump.fmt = self.__format_name__

        if fmt is None:
            if dump.fmt:
                dump.fmt = f"{dump.fmt}:"

            if isinstance(value, (bytes, bytearray)):
                dump.fmt += "bytes"
                bytes_x.__pack__(value, pieces, dump)

            elif isinstance(value, tuple):
                try:
                    item, bound_fmt = value
                except ValueError:
                    dump.value = value
                    dump.fmt += "unknown"
                    raise ValueError(
                        f"expected (value, fmt) pair, but got tuple of length {len(value)}"
                    ) from None
                dump.fmt = bound_fmt
                self.__pack_and_dump__(item, pieces, dump, bound_fmt)

            elif isinstance(value, Data):
                dump.fmt += type(value).name
                value.__pack__(value, pieces, dump)

            elif isinstance(value, list):
                dump.fmt += "list"
                for i, item in enumerate(value):
                    self.__pack_and_dump__(
                        item, pieces, dump.add_record(access=f"[{i}]"), fmt=None
                    )

            elif isinstance(value, dict):
                dump.fmt += "dict"
                for name, item in value.items():
                    self.__pack_and_dump__(
                        item, pieces, dump.add_record(access=f"[{name!r}]"), fmt=None
                    )

            else:
                dump.fmt += type(value).__name__ + " (invalid)"
                dump.value = value

                raise TypeError(
                    "no format specified, value must be a packable "
                    "data type or (value, fmt) pairing (or dict/list of them)"
                )

        elif isinstance(fmt, dict):
            if not isinstance(value, dict):
                dump.fmt = "dict"
                dump.value = value
                raise TypeError(f"invalid value, expected dict, got {value!r}")

            missing = []

            for name, subfmt in fmt.items():
                subdump = dump.add_record(access=f"[{name!r}]")

                if isinstance(subfmt, (DataMeta, Transform)):
                    subdump.fmt = subfmt.name

                try:
                    subvalue = value[name]
                except KeyError:
                    missing.append(name)
                    subdump.value = "(missing)"
                    continue

                if isinstance(subfmt, (DataMeta, Transform)):
                    subfmt.__pack__(subvalue, pieces, subdump)
                else:
                    # must be dict, list, tuple
                    self.__pack_and_dump__(subvalue, pieces, subdump, subfmt)

            if missing:
                if len(missing) > 1:
                    missing_values = f"missing values: {repr(missing)[1:-1]}"
                else:
                    missing_values = f"missing value: {missing[0]!r}"
                raise TypeError(missing_values)

            extra_keys = set(value) - set(fmt)

            if extra_keys:
                separate = True
                for name in extra_keys:
                    dump.add_record(
                        access=f"[{name!r}]",
                        value=value[name],
                        fmt="(unexpected)",
                        separate=separate,
                    )
                    separate = False

                value_s = "value" if len(extra_keys) == 1 else "values"
                names = ", ".join(repr(name) for name in extra_keys)
                raise TypeError(f"unexpected {value_s}: {names}")

        elif isinstance(fmt, (tuple, list)):
            if not isinstance(value, (tuple, list)):
                dump.fmt = "tuple or list"
                dump.value = value
                raise TypeError(f"invalid value, expected list or tuple, got {value!r}")

            for i, (subfmt, item) in enumerate(zip(fmt, value)):
                subdump = dump.add_record(access=f"[{i}]")
                if isinstance(subfmt, (DataMeta, Transform)):
                    subdump.fmt = subfmt.name
                    subfmt.__pack__(item, pieces, subdump)
                else:
                    # must be dict, list, tuple, None
                    self.__pack_and_dump__(item, pieces, subdump, subfmt)

            if len(value) != len(fmt):
                if len(value) < len(fmt):
                    for i in range(len(value), len(fmt)):
                        dump.add_record(fmt=fmt[i], value="(missing)", access=f"[{i}]")
                else:
                    separate = True
                    for i in range(len(fmt), len(value)):
                        dump.add_record(
                            fmt="(unexpected)",
                            value=value[i],
                            access=f"[{i}]",
                            separate=separate,
                        )
                        separate = False

                value_s = "value" if len(value) == 1 else "values"
                raise TypeError(f"{len(value)} {value_s} given, expected {len(fmt)}")

        elif isinstance(fmt, (DataMeta, Transform)):
            dump.fmt = fmt.name
            fmt.__pack__(value, pieces, dump)
        else:
            if isinstance(fmt, type):
                dump.fmt = fmt.__name__ + " (invalid)"
            else:
                dump.fmt = str(fmt) + " (invalid)"

            dump.value = value

            raise TypeError(
                "bad item format, must be a packable data type/transform "
                "(or a dict, list, or tuple of them)"
            )

    def __unpack__(
        self,
        buffer: bytes,
        offset: int,
        dump: Optional[Record] = None,
        fmt: DefaultFormat = Default.UNSPECIFIED,
    ) -> Tuple[Any, int]:
        # pylint: disable=arguments-differ
        if dump is not None:
            return self.__unpack_and_dump__(buffer, offset, dump, fmt)

        if fmt is Default.UNSPECIFIED:
            fmt = self._fmt

        if fmt is None:
            return bytes_x.__unpack__(buffer, offset, dump)

        if isinstance(fmt, (list, tuple)):
            list_value = []
            for subfmt in fmt:
                item, offset = self.__unpack__(buffer, offset, dump, subfmt)
                list_value.append(item)

            if isinstance(fmt, tuple):
                return tuple(list_value), offset

            return list_value, offset

        if isinstance(fmt, dict):
            dict_value = {}
            for name, subfmt in fmt.items():
                item, offset = self.__unpack__(buffer, offset, dump, subfmt)
                dict_value[name] = item

            return dict_value, offset

        return fmt.__unpack__(buffer, offset)

    def __unpack_and_dump__(
        self,
        buffer: bytes,
        offset: int,
        dump: Record,
        fmt: DefaultFormat = Default.UNSPECIFIED,
    ) -> Tuple[Any, int]:
        # pylint: disable=arguments-differ
        if fmt is Default.UNSPECIFIED:
            fmt = self._fmt
            dump.fmt = self.__format_name__

        if fmt is None:
            if dump.fmt:
                dump.fmt = f"{dump.fmt}:bytes"
            else:
                dump.fmt = "bytes"

            return bytes_x.__unpack__(buffer, offset, dump)

        if isinstance(fmt, (list, tuple)):
            list_value = []
            for i, subfmt in enumerate(fmt):
                item, offset = self.__unpack_and_dump__(
                    buffer, offset, dump.add_record(access=f"[{i}]"), subfmt
                )
                list_value.append(item)

            if isinstance(fmt, tuple):
                return tuple(list_value), offset

            return list_value, offset

        if isinstance(fmt, dict):
            dict_value = {}
            for name, subfmt in fmt.items():
                item, offset = self.__unpack_and_dump__(
                    buffer, offset, dump.add_record(access=f"[{name!r}]"), subfmt
                )
                dict_value[name] = item

            return dict_value, offset

        if isinstance(fmt, (DataMeta, Transform)):
            dump.fmt = fmt.name
            return fmt.__unpack__(buffer, offset, dump)

        dump.fmt = str(fmt)

        raise TypeError(
            "bad item format, must be a packable data type/transform "
            "(or a dict, list, or tuple of them)"
        )


items = ItemsX(name="items")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test calcsize() utility function."""

import pytest
from baseline import Baseline

from plum.array import ArrayX
from plum.exceptions import SizeError
from plum.littleendian import uint8, uint16
from plum.structure import Structure, member
from plum.utilities import calcsize

greedy_array = ArrayX(name="greedy_array", fmt=uint8)


class TestFixed:
    class Struct(Structure):
        m0 = member(fmt=uint8)

    def test_transform(self):
        assert calcsize(uint8) == 1

    def test_data_store_type(self):
        assert calcsize(self.Struct) == 1

    def test_data_store_instance(self):
        assert calcsize(self.Struct(m0=0)) == 1

    def test_dict(self):
        assert calcsize({"a": uint8, "b": uint16}) == 3

    def test_list(self):
        assert calcsize([uint8, uint16]) == 3

    def test_tuple(self):
        assert calcsize((uint8, uint16)) == 3


class TestVariable:
    class Struct(Structure):
        m0 = member(fmt=greedy_array)

    expected_message = Baseline(
        """
        size varies
        """
    )

    def test_transform(self):
        expected_message = Baseline(
            """
            <transform 'greedy_array'> sizes vary
            """
        )
        with pytest.raises(SizeError) as trap:
            calcsize(greedy_array)
        assert str(trap.value) == expected_message

    def test_data_store_type(self):
        expected_message = Baseline(
            """
            size varies
            """
        )
        with pytest.raises(SizeError) as trap:
            calcsize(calcsize(self.Struct))
        assert str(trap.value) == expected_message

    def test_dict(self):
        with pytest.raises(SizeError) as trap:
            calcsize({"a": uint8, "b": None})
        assert str(trap.value) == self.expected_message

    def test_list(self):
        with pytest.raises(SizeError) as trap:
            calcsize([uint8, None])
        assert str(trap.value) == self.expected_message

    def test_tuple(self):
        with pytest.raises(SizeError) as trap:
            calcsize((uint8, None))
        assert str(trap.value) == self.expected_message

    def test_none(self):
        with pytest.raises(SizeError) as trap:
            calcsize(None)
        assert str(trap.value) == self.expected_message


class TestInvalid:
    def test_invalid(self):
        expected_message = Baseline(
            """
            invalid format <class 'int'>
            """
        )

        with pytest.raises(TypeError) as trap:
            calcsize(int)

        assert str(trap.value) == expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Classes and utilities for packing/unpacking bytes."""

import os

enable_boost = os.environ.get("ENABLE_PLUM_BOOST", "AUTO").upper()

"""
if enable_boost in {"AUTO", "YES"}:  # pragma: no cover
    try:
        import plum_boost as boost  # pylint: disable=import-error
    except ModuleNotFoundError:
        if enable_boost == "YES":
            raise
        boost = None
    else:
        pack = data.plum_namespace["pack"] = boost.pack
        unpack = data.plum_namespace["unpack"] = boost.unpack
        data.Data.pack = boost.PackMethod()
        data.Data.unpack = classmethod(boost.unpack)

elif enable_boost == "NO":  # pragma: no cover
    boost = None

else:  # pragma: no cover
    raise RuntimeError(
        "ENABLE_PLUM_BOOST environment variable must be YES, NO, or AUTO"
    )
"""

boost = None

del enable_boost, os



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test array transform properties."""

import pytest

from plum.array import ArrayX
from plum.exceptions import SizeError
from plum.littleendian import single, uint8
from plum.str import StrX

ascii_zero = StrX(name="ascii_zero", encoding="ascii", zero_termination=True)


class TestDefault:

    """Test with as many left to default as possible."""

    arrayx = ArrayX(uint8)

    def test_name(self):
        assert self.arrayx.name == "List[int]"

    def test_hint(self):
        assert self.arrayx.__hint__ == "List[int]"

    def test_nbytes(self):
        with pytest.raises(SizeError):
            return self.arrayx.nbytes

    def test_fmt(self):
        assert self.arrayx.fmt is uint8

    def test_dims(self):
        assert self.arrayx.dims == (None,)


class TestPositional:

    """Test explicitly defined with positional argument."""

    arrayx = ArrayX(uint8, (2,), "name")

    def test_name(self):
        assert self.arrayx.name == "name"

    def test_hint(self):
        assert self.arrayx.__hint__ == "List[int]"

    def test_nbytes(self):
        assert self.arrayx.nbytes == 2

    def test_fmt(self):
        assert self.arrayx.fmt is uint8

    def test_dims(self):
        assert self.arrayx.dims == (2,)


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    arrayx = ArrayX(fmt=uint8, dims=(2,), name="name")


class TestVariable:

    """Test nbytes property when member size is variable."""

    arrayx = ArrayX(name="arrayx", fmt=ascii_zero, dims=(2,))

    def test_nbytes(self):
        with pytest.raises(SizeError):
            return self.arrayx.nbytes


class TestDefaultNameHint:

    """Test generation of name and type hint."""

    def test_multi_dims(self):
        arrayx = ArrayX(fmt=uint8, dims=[20, 30, None])
        assert arrayx.name == arrayx.__hint__ == "List[List[List[int]]]"

    def test_greedy_dim(self):
        arrayx = ArrayX(fmt=single)
        assert arrayx.name == arrayx.__hint__ == "List[float]"

    def test_variable_single_dim(self):
        arrayx = ArrayX(fmt=single, dims=uint8)
        assert arrayx.name == arrayx.__hint__ == "List[float]"

    def test_variable_multi_dim(self):
        dimsx = ArrayX(fmt=uint8, dims=[3])
        arrayx = ArrayX(fmt=single, dims=dimsx)
        assert arrayx.name == arrayx.__hint__ == "List[List[List[float]]]"

    def test_name_override(self):
        arrayx = ArrayX(fmt=single, name="array")
        assert arrayx.__hint__ == "List[float]"
        assert arrayx.name == "array"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test unpack() utility function."""

from typing import Optional, Tuple

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.data import Data, DataMeta
from plum.dump import Record
from plum.exceptions import InsufficientMemoryError, ExcessMemoryError, UnpackError
from plum.littleendian import uint8, uint16
from plum.utilities import getbytes, unpack, unpack_and_dump
from plum.structure import Structure, member


class CustomError(Data, metaclass=DataMeta):  # pylint: disable=abstract-method

    """Always raises ValueError when unpacked."""

    __hint__ = "CustomError"
    __format_name__ = "CustomError (Structure)"

    @classmethod
    def __unpack__(
        cls, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[int, int]:
        if dump is not None:
            dump.value = "<invalid>"
            getbytes(buffer, offset, dump, 1)
        raise ValueError("invalid byte")


class CustomStructure(Structure):

    """Has member that raises ValueError when unpacked."""

    m1 = member(fmt=uint16)
    m2 = member(fmt=CustomError)


class TestFormatVariations:

    """Test format variations."""

    def test_plum_type(self):
        """Test format is a plum type."""
        exp_dump = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            | 0      | 258   | 02 01 | uint16 |
            +--------+-------+-------+--------+
            """
        )

        buffer = b"\x02\x01"
        fmt = uint16

        item = unpack(fmt, buffer)
        assert item == 0x0102

        item, dump = unpack_and_dump(fmt, buffer)
        assert item == 0x0102
        assert str(dump) == exp_dump

    def test_tuple(self):
        """Test format is a tuple of plum types."""
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            | 0      | [0]    | 2     | 02    | uint8  |
            | 1      | [1]    | 1     | 01    | uint8  |
            +--------+--------+-------+-------+--------+
            """
        )

        buffer = b"\x02\x01"
        fmt = (uint8, uint8)

        item = unpack(fmt, buffer)
        assert item == (2, 1)

        item, dump = unpack_and_dump(fmt, buffer)
        assert item == (2, 1)
        assert str(dump) == exp_dump

    def test_list(self):
        """Test format is a list of plum types."""
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            | 0      | [0]    | 2     | 02    | uint8  |
            | 1      | [1]    | 1     | 01    | uint8  |
            +--------+--------+-------+-------+--------+
            """
        )

        buffer = b"\x02\x01"
        fmt = [uint8, uint8]

        item = unpack(fmt, buffer)
        assert item == [2, 1]

        item, dump = unpack_and_dump(fmt, buffer)
        assert item == [2, 1]
        assert str(dump) == exp_dump

    def test_dict(self):
        """Test format is a dictionary of plum types."""
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            | 0      | ['a']  | 2     | 02    | uint8  |
            | 1      | ['b']  | 1     | 01    | uint8  |
            +--------+--------+-------+-------+--------+
            """
        )

        buffer = b"\x02\x01"
        fmt = {"a": uint8, "b": uint8}

        item = unpack(fmt, buffer)
        assert item == {"a": 2, "b": 1}

        item, dump = unpack_and_dump(fmt, buffer)
        assert item == {"a": 2, "b": 1}
        assert str(dump) == exp_dump

    def test_nested(self):
        """Test format is a dictionary of plum types."""
        exp_dump = Baseline(
            """
            +--------+----------+-------+-------+--------+
            | Offset | Access   | Value | Bytes | Format |
            +--------+----------+-------+-------+--------+
            |        | ['a']    |       |       |        |
            | 0      |   [0]    | 1     | 01    | uint8  |
            | 1      |   [1]    | 2     | 02    | uint8  |
            |        | ['b']    |       |       |        |
            | 2      |   ['i']  | 3     | 03    | uint8  |
            | 3      |   ['ii'] | 4     | 04    | uint8  |
            | 4      | ['c']    | 5     | 05    | uint8  |
            +--------+----------+-------+-------+--------+
            """
        )

        buffer = b"\x01\x02\x03\x04\x05"
        fmt = {"a": (uint8, uint8), "b": {"i": uint8, "ii": uint8}, "c": uint8}

        item = unpack(fmt, buffer)
        assert item == {"a": (1, 2), "b": {"i": 3, "ii": 4}, "c": 5}

        item, dump = unpack_and_dump(fmt, buffer)
        assert item == {"a": (1, 2), "b": {"i": 3, "ii": 4}, "c": 5}
        assert str(dump) == exp_dump


class TestExceptions:

    """Test various exception paths."""

    def test_value_error(self):
        """Verify UnpackError raised with context of original unpacking exception."""
        exp_message = Baseline(
            """
            +--------+--------+-----------+-------+-----------------------------+
            | Offset | Access | Value     | Bytes | Format                      |
            +--------+--------+-----------+-------+-----------------------------+
            |        |        |           |       | CustomStructure (Structure) |
            | 0      | m1     | 258       | 02 01 | uint16                      |
            | 2      | m2     | <invalid> | 00    | CustomError (Structure)     |
            +--------+--------+-----------+-------+-----------------------------+

            ValueError occurred during unpack operation:

            invalid byte
            """
        )
        fmt = CustomStructure
        buffer = b"\x02\x01\x00"

        with pytest.raises(UnpackError) as trap:
            unpack(fmt, buffer)
        assert wrap_message(trap.value) == exp_message
        assert isinstance(trap.value.__context__, ValueError)

        with pytest.raises(UnpackError) as trap:
            unpack_and_dump(fmt, buffer)
        assert wrap_message(trap.value) == exp_message
        assert isinstance(trap.value.__context__, ValueError)


class TestUnpackBytes:

    """Test buffer argument as "bytes" type."""

    # FUTURE - move this to conformance test

    fmt = uint16
    buffer_type = bytes
    buffer_data = b"\x02\x01"
    offset = 0

    exp_value = 0x0102
    exp_dump = Baseline(
        """
        +--------+-------+-------+--------+
        | Offset | Value | Bytes | Format |
        +--------+-------+-------+--------+
        | 0      | 258   | 02 01 | uint16 |
        +--------+-------+-------+--------+
        """
    )

    def test_sample(self):
        """Verify unpack successful."""

        buffer = self.buffer_type(self.buffer_data)
        assert unpack(self.fmt, buffer) == self.exp_value

        buffer = self.buffer_type(self.buffer_data)
        value, dump = unpack_and_dump(self.fmt, buffer)
        assert (value, str(dump)) == (self.exp_value, self.exp_dump)

    exp_insufficient_message = Baseline(
        """
        +--------+----------------------+-------+--------+
        | Offset | Value                | Bytes | Format |
        +--------+----------------------+-------+--------+
        | 0      | <insufficient bytes> | 02    | uint16 |
        +--------+----------------------+-------+--------+

        InsufficientMemoryError occurred during unpack operation:

        1 too few bytes to unpack uint16, 2 needed, only 1 available
        """
    )

    def test_insufficient(self):
        """Verify UnpackError with InsufficientMemoryError context when too few bytes."""
        buffer = self.buffer_type(self.buffer_data[:-1])
        with pytest.raises(UnpackError) as trap:
            unpack(self.fmt, buffer)
        assert wrap_message(trap.value) == self.exp_insufficient_message
        assert isinstance(trap.value.__context__, InsufficientMemoryError)

        buffer = self.buffer_type(self.buffer_data[:-1])
        with pytest.raises(UnpackError) as trap:
            unpack_and_dump(self.fmt, buffer)
        assert wrap_message(trap.value) == self.exp_insufficient_message
        assert isinstance(trap.value.__context__, InsufficientMemoryError)

    exp_excess_message = Baseline(
        """
            +--------+----------------+-------------------------------------------------+--------+
            | Offset | Value          | Bytes                                           | Format |
            +--------+----------------+-------------------------------------------------+--------+
            |  0     | 258            | 02 01                                           | uint16 |
            +--------+----------------+-------------------------------------------------+--------+
            |  2     | <excess bytes> | 99 9a 9b 9c 9d 9e 9f a0 a1 a2 a3 a4 a5 a6 a7 a8 |        |
            | 18     |                | a9 aa ab ac ad ae af b0 b1 b2 b3 b4             |        |
            +--------+----------------+-------------------------------------------------+--------+

            ExcessMemoryError occurred during unpack operation:

            28 unconsumed bytes
            """
    )

    def test_excess(self):
        """Verify UnpackError with ExcessMemoryError context when too many bytes."""
        excess = bytes(range(0x99, 0xB5))
        buffer = self.buffer_type(self.buffer_data + excess)
        with pytest.raises(UnpackError) as trap:
            unpack(self.fmt, buffer)
        assert wrap_message(trap.value) == self.exp_excess_message
        assert isinstance(trap.value.__context__, ExcessMemoryError)

        buffer = self.buffer_type(self.buffer_data + excess)
        with pytest.raises(UnpackError) as trap:
            unpack_and_dump(self.fmt, buffer)
        assert wrap_message(trap.value) == self.exp_excess_message
        assert isinstance(trap.value.__context__, ExcessMemoryError)


class TestUnpackByteArray(TestUnpackBytes):

    """Test buffer argument as "bytearray" type."""

    buffer_type = bytearray



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test use of BitFields as a typ in bitfield_member()."""

# pylint: disable=too-many-lines

from baseline import Baseline

from plum.bigendian import uint8
from plum.bitfields import BitFields, bitfield
from plum.conformance import Case, CaseData
from plum.structure import Structure, member, bitfield_member


class Bits(BitFields, nbytes=1):

    b1: int = bitfield(size=1, typ=bool)
    b2: int = bitfield(size=1, typ=bool)


class Test(Case):

    """bitfield_member(typ=BitFields)."""

    class MyStruct(Structure):

        m1: Bits = bitfield_member(typ=Bits, size=2)
        m2: int = bitfield_member(typ=int, size=6)
        m3: int = member(fmt=uint8, default=0xFF)

    data = CaseData(
        fmt=MyStruct,
        bindata=b"\x80\xff",
        nbytes=2,
        values=(
            MyStruct(m1=Bits.from_int(2), m2=0, m3=0xFF),
            [2, 0, 0xFF],
        ),
        dump=Baseline(
            """
            +--------+--------+-------+----------+----------------------+
            | Offset | Access | Value | Bytes    | Format               |
            +--------+--------+-------+----------+----------------------+
            |        |        |       |          | MyStruct (Structure) |
            | 0      |        | 128   | 80       |                      |
            |        | m1     | 2     |          | Bits (BitFields)     |
            |  [7]   |   b1   | True  | 1....... | bool                 |
            |  [6]   |   b2   | False | .0...... | bool                 |
            |  [0:6] | m2     | 0     | ..000000 | int                  |
            | 1      | m3     | 255   | ff       | uint8                |
            +--------+--------+-------+----------+----------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+--------+----------------+----------+----------------------+
            | Offset | Access | Value          | Bytes    | Format               |
            +--------+--------+----------------+----------+----------------------+
            |        |        |                |          | MyStruct (Structure) |
            | 0      |        | 128            | 80       |                      |
            |        | m1     | 2              |          | Bits (BitFields)     |
            |  [7]   |   b1   | True           | 1....... | bool                 |
            |  [6]   |   b2   | False          | .0...... | bool                 |
            |  [0:6] | m2     | 0              | ..000000 | int                  |
            | 1      | m3     | 255            | ff       | uint8                |
            +--------+--------+----------------+----------+----------------------+
            | 2      |        | <excess bytes> | 99       |                      |
            +--------+--------+----------------+----------+----------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+--------+----------------------+----------+----------------------+
            | Offset | Access | Value                | Bytes    | Format               |
            +--------+--------+----------------------+----------+----------------------+
            |        |        |                      |          | MyStruct (Structure) |
            | 0      |        | 128                  | 80       |                      |
            |        | m1     | 2                    |          | Bits (BitFields)     |
            |  [7]   |   b1   | True                 | 1....... | bool                 |
            |  [6]   |   b2   | False                | .0...... | bool                 |
            |  [0:6] | m2     | 0                    | ..000000 | int                  |
            |        | m3     | <insufficient bytes> |          | uint8                |
            +--------+--------+----------------------+----------+----------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack uint8, 1 needed, only 0 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, m1: 'Bits', m2: int, m3: Optional[int] = None) -> None:
                if m3 is None:
                    m3 = type(self).m3.default

                if not 0 <= m1 <= 3:
                   raise ValueError("'m1' out of range, 0 <= m1 <= 3")

                if not 0 <= m2 <= 63:
                   raise ValueError("'m2' out of range, 0 <= m2 <= 63")

                self[:] = (m1, m2, m3)

            @m1.getter
            def m1(self) -> 'Bits':
                return self[0]

            @m1.setter
            def m1(self, value: 'Bits') -> None:
                if not 0 <= value <= 3:
                   raise ValueError("out of range, 0 <= value <= 3")
                self[0] = type(self).m1.typ(value)

            @m2.getter
            def m2(self) -> int:
                return self[1]

            @m2.setter
            def m2(self, value: int) -> None:
                if not 0 <= value <= 63:
                   raise ValueError("out of range, 0 <= value <= 63")
                self[1] = value

            @m3.getter
            def m3(self) -> int:
                return self[2]

            @m3.setter
            def m3(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_m1, m_m2, m_m3) = value

                if dump is None:
                    bitfields_0 = 0

                    bitfields_0 |= (bitfields_0 & -193) | ((m_m1 & 0x3) << 6)

                    bitfields_0 |= (bitfields_0 & -64) | (m_m2 & 0x3f)

                    pieces.append(bitfields_0.to_bytes(1, "little", signed=False))

                    cls.m3.fmt.__pack__(m_m3, pieces, dump)

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0 = 0

                    dump_m_m1 = dump.add_record(access="m1", value=int(m_m1), fmt=cls.m1.typ)
                    cls.m1.typ.__add_bitfields_to_dump__(m_m1, dump_m_m1, bitoffset=6)
                    bitfields_0 |= (bitfields_0 & -193) | ((m_m1 & 0x3) << 6)

                    dump.add_record(access="m2", bits=(0, 6), value=m_m2, fmt=int)
                    bitfields_0 |= (bitfields_0 & -64) | (m_m2 & 0x3f)

                    pieces.append(bitfields_0.to_bytes(1, "little", signed=False))
                    bitfields_0_dump.value = str(bitfields_0)
                    bitfields_0_dump.memory = pieces[-1]

                    m3_dump = dump.add_record(access="m3", fmt=cls.m3.fmt)
                    cls.m3.fmt.__pack__(m_m3, pieces, m3_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["MyStruct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, dump, "little", nbytes=1, signed=False)

                    m_m1 = (bitfields_0 >> 6) & 0x3
                    m_m1 = cls.m1.typ.from_int(m_m1)

                    m_m2 = (bitfields_0 >> 0) & 0x3f

                    m_m3, offset = cls.m3.fmt.__unpack__(buffer, offset, dump)

                else:
                    bitfields_0_dump = dump.add_record()
                    bitfields_0, offset = cls.m1.__unpack__(buffer, offset, bitfields_0_dump, "little", nbytes=1, signed=False)

                    m_m1 = (bitfields_0 >> 6) & 0x3
                    m_m1 = cls.m1.typ.from_int(m_m1)
                    dump_m_m1 = dump.add_record(access="m1", value=int(m_m1), fmt=cls.m1.typ)
                    cls.m1.typ.__add_bitfields_to_dump__(m_m1, dump_m_m1, bitoffset=6)

                    m_m2 = (bitfields_0 >> 0) & 0x3f
                    dump.add_record(access="m2", bits=(0, 6), value=m_m2, fmt=int)

                    m3_dump = dump.add_record(access="m3", fmt=cls.m3.fmt)
                    m_m3, offset = cls.m3.fmt.__unpack__(buffer, offset, m3_dump)

                structure[:] = (m_m1, m_m2, m_m3)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(m1={self.m1!r}, m2={self.m2!r}, m3={self.m3!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Utility functions."""

from typing import Tuple, Union

from .dump import Record
from .exceptions import InsufficientMemoryError


def getbytes(
    buffer: bytes, offset: int, dump: Record, nbytes: Union[int, None] = None
) -> Tuple[bytes, int]:
    """Get bytes from buffer (get remainder in buffer if nbytes is None)."""
    if nbytes is None:
        chunk = bytes(buffer[offset:])

        offset += len(chunk)

        dump.memory = chunk

    else:
        start = offset
        offset += nbytes
        chunk = bytes(buffer[start:offset])

        if len(chunk) < nbytes:
            dump.value = "<insufficient bytes>"
            if len(chunk) > 16:
                dump.add_extra_bytes("", chunk)
            else:
                dump.memory = chunk

            extra = f" {dump.fmt}" if dump.fmt else ""

            unpack_shortage = (
                f"{nbytes - len(chunk)} too few bytes to unpack{extra}"
                f", {nbytes} needed, only {len(chunk)} available"
            )

            raise InsufficientMemoryError(unpack_shortage)

        dump.memory = chunk

    return chunk, offset



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Verify accuracy of example code in all docstrings and RST files."""

import doctest
import os
import sys

import pytest


def get_docfiles():
    """Walk repository and return list of all .rst and .py files.

    :returns: list of all .py and .rst files
    :rtype: list of str

    """
    root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    docfiles = []

    for subdir in ["docs", "src", "boost"]:
        for dirpath, _dirnames, filenames in os.walk(os.path.join(root, subdir)):
            if os.path.join("plum-py", "build") in dirpath:
                continue
            for filename in filenames:
                if os.path.splitext(filename)[1] not in {".rst", ".py"}:
                    continue
                if ".tox" in dirpath:
                    continue  # do not test examples from 3rd-party packages
                if filename in {
                    "about.rst",  # contains examples of future functionality
                    "custommethods.rst",  # contains code injection examples
                }:
                    continue
                if sys.version_info < (3, 11) and filename == "flag.rst":
                    continue  # 3.11 reversed order of repr of or'd flags
                docfiles.append(os.path.join(dirpath, filename))
    return docfiles


OPTION_FLAGS = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE


@pytest.mark.parametrize("docfile", get_docfiles())
def test_examples(docfile: str) -> None:
    """Test interactive examples in documentation file.

    :param docfile: path of RST file

    """
    failure_count, _test_count = doctest.testfile(
        docfile, module_relative=False, optionflags=OPTION_FLAGS
    )

    assert failure_count == 0



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test NoneX transform exceptions."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import PackError
from plum.none import NoneX

nonex = NoneX(name="nonex")


class TestInvalidPackValue:

    expected_message = Baseline(
        """
        +--------+-------+-------+--------+
        | Offset | Value | Bytes | Format |
        +--------+-------+-------+--------+
        |        | 0     |       | nonex  |
        +--------+-------+-------+--------+

        TypeError occurred during pack operation:

        value must be 'None'
        """
    )

    def test_pack(self):
        with pytest.raises(PackError) as trap:
            nonex.pack(0)

        assert wrap_message(trap.value) == self.expected_message
        assert isinstance(trap.value.__context__, TypeError)

    def test_nbytes(self):
        with pytest.raises(PackError) as trap:
            nonex.pack_and_dump(0)

        assert wrap_message(trap.value) == self.expected_message
        assert isinstance(trap.value.__context__, TypeError)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structure type view."""

from collections.abc import Sequence

from ..view import PlumView

# Attributes __type__, __buffer__, and __offset__ assigned in __init__ of PlumView.
# pylint: disable=no-member


class StructureView(PlumView, Sequence):  # pylint: disable=too-many-ancestors

    """Structure type view."""

    def __add__(self, other):
        return self.unpack() + other

    def append(self, item):
        """Append object to the end of the list."""
        raise TypeError(f"{self.__class__.__name__!r} does not support append()")

    def asdict(self):
        """Return structure members in dictionary form.

        :returns: structure members
        :rtype: dict

        """
        return self.unpack().asdict()

    def clear(self):
        """Remove all items from list."""
        raise TypeError(f"{self.__class__.__name__!r} does not support clear()")

    def copy(self):
        """Return a shallow copy of the list.

        :returns: shallow copy
        :rtype: list

        """
        return self.unpack().copy()

    def count(self, item):  # pylint: disable=arguments-differ
        """Return number of occurrences of value.

        :returns: number of occurrences
        :type: int

        """
        return self.unpack().count(item)

    def __delattr__(self, item):
        raise TypeError(
            f"{self.__class__.__name__!r} does not support attribute deletion"
        )

    def __delitem__(self, key):
        raise TypeError(f"{self.__class__.__name__!r} does not support item deletion")

    def __eq__(self, other):
        return self.unpack() == other

    def extend(self, item):
        """Extend list by appending elements from the iterable."""
        raise TypeError(f"{self.__class__.__name__!r} does not support extend()")

    def __ge__(self, other):
        return self.unpack() >= other

    def __getattr__(self, name):
        member_names, _member_types = self.__type__.__names_types__
        try:
            index = member_names.index(name)
        except ValueError:
            # let normal Python AttributeError exception occur
            object.__getattribute__(self, name)

        return self.__getitem__(index)

    def __getitem__(self, index):
        _names, types = self.__type__.__names_types__
        return types[index].view(
            self.__buffer__, offset=self.__type__.__offsets__[index] + self.__offset__
        )

    def __gt__(self, other):
        return self.unpack() > other

    def __iadd__(self, other):
        raise TypeError(
            f"{self.__class__.__name__!r} does not support in-place addition"
        )

    def __imul__(self, other):
        raise TypeError(
            f"{self.__class__.__name__!r} does not support in-place multiplication"
        )

    def index(self, item):  # pylint: disable=arguments-differ
        """Return first index of value.

        :returns: first index of value
        :type: int

        """
        return self.unpack().index(item)

    def insert(self, item, index):
        """Insert object before index."""
        raise TypeError(f"{self.__class__.__name__!r} does not support insert()")

    def __le__(self, other):
        return self.unpack() <= other

    def __len__(self):
        return len(self.__type__.__names_types__)

    def __lt__(self, other):
        return self.unpack() < other

    def __mul__(self, other):
        return self.unpack() * other

    def pop(self, index=-1):
        """Remove and return item at index (default last)."""
        raise TypeError(f"{self.__class__.__name__!r} does not support pop()")

    def __radd__(self, other):
        return other + self.unpack()

    def remove(self, item):
        """Remove first occurrence of value."""
        raise TypeError(f"{self.__class__.__name__!r} does not support remove()")

    def reverse(self):
        """Reverse in place."""
        raise TypeError(f"{self.__class__.__name__!r} does not support reverse()")

    def __rmul__(self, other):
        return other * self.unpack()

    def __setattr__(self, key, value):
        names, _types = self.__type__.__names_types__
        if key in names:
            # Attribute name is a structure member, set value of the corresponding view.
            self.__getattr__(key).set(value)
        else:
            # Attribute name is not a structure member, set attribute of this view.
            object.__setattr__(self, key, value)

    def __setitem__(self, index, value):
        self.__getitem__(index).set(value)

    def sort(self):
        """Stable sort in place."""
        self.set(sorted(self.unpack()))



"""Test decimal transform properties."""

from plum.decimal import DecimalX


class TestProperties:

    """Test with as many left to default as possible."""

    u16p1 = DecimalX(2, 1, "big", signed=False, name="u16p1")

    def test_hint(self):
        assert self.u16p1.__hint__ == "Decimal"

    def test_name(self):
        assert self.u16p1.name == "u16p1"

    def test_nbytes(self):
        assert self.u16p1.nbytes == 2

    def test_byteorder(self):
        assert self.u16p1.byteorder == "big"

    def test_signed(self):
        assert self.u16p1.signed is False

    def test_precision(self):
        assert self.u16p1.precision == 1


class TestDefaultNameAndHint:
    def test_hint(self):
        xform = DecimalX(2, 1)
        assert xform.__hint__ == "Decimal"

    def test_name(self):
        xform = DecimalX(2, 1)
        assert xform.name == "Decimal (precision=1)"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Sample integer flag enumeration"""

from enum import IntFlag, Flag


class Register(IntFlag):

    """Sample integer flag enumeration."""

    SP = 1
    R0 = 2
    R1 = 4


class RegisterMixin(int, Flag):

    """Sample integer flag mixin."""

    SP = 1
    R0 = 2
    R1 = 4



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test operations int view of buffer bytes."""

from math import ceil, floor, trunc

import pytest
from baseline import Baseline

from plum.bigendian import uint8, uint16
from plum.conformance import wrap_message
from plum.utilities import pack


# pylint: disable=redefined-outer-name
@pytest.fixture()
def uint8_view():
    """Sample uint8 plum view."""
    buffer = b"\x01\x02\x0A\x0B"
    yield uint8.view(buffer, offset=2)  # decimal value of 10


class TestIntView:

    """Test operations int view of buffer bytes."""

    def test_arithmetic_assignments(self):
        """Test augmented arithmetic assignment operations."""
        memory_backing = bytearray([0x00, 0x01, 0x04, 0xFF])
        uint16_view = uint16.view(memory_backing, offset=1)  # decimal value of 260

        # __iadd__
        uint16_view += 8
        assert uint16_view == 268
        assert memory_backing == bytearray(b"\x00\x01\x0c\xff")

        # __isub__
        uint16_view -= 204
        assert uint16_view == 64
        assert memory_backing == bytearray(b"\x00\x00\x40\xff")

        # __imul__
        uint16_view *= 16
        assert uint16_view == 1024
        assert memory_backing == bytearray(b"\x00\x04\x00\xff")

        # __imod__
        uint16_view %= 306
        assert uint16_view == 106
        assert memory_backing == bytearray(b"\x00\x00\x6a\xff")

        # __ipow__
        uint16_view **= 2
        assert uint16_view == 11236
        assert memory_backing == bytearray(b"\x00\x2b\xe4\xff")

        # __ilshift__
        uint16_view <<= 1
        assert uint16_view == 22472
        assert memory_backing == bytearray(b"\x00\x57\xc8\xff")

        # __irshift__
        uint16_view >>= 8
        assert uint16_view == 87
        assert memory_backing == bytearray(b"\x00\x00\x57\xff")

        # __ixor__
        uint16_view ^= 0x0F0F
        assert uint16_view == 3928
        assert memory_backing == bytearray(b"\x00\x0f\x58\xff")

        # __iand__
        uint16_view &= 0xFF08
        assert uint16_view == 3848
        assert memory_backing == bytearray(b"\x00\x0f\x08\xff")

        # __ior__
        uint16_view |= 0xA0F0
        assert uint16_view == 45048
        assert memory_backing == bytearray(b"\x00\xaf\xf8\xff")

    def test_bitwise(self, uint8_view):
        """Test bitwise operations."""
        # __lshift__ and __rlshift__
        assert uint8_view << 2 == 40
        assert 2 << uint8_view == 2048

        # __rshift__ and __rrshift__
        assert uint8_view >> 1 == 5
        assert 4096 >> uint8_view == 4

        # __and__ and __rand__
        assert uint8_view & 3 == 2
        assert 12 & uint8_view == 8

        # __xor__ and __rxor__
        assert uint8_view ^ 8 == 2
        assert 14 ^ uint8_view == 4

        # __or__ and __ror__
        assert uint8_view | 7 == 15
        assert 3 | uint8_view == 11

    def test_comparisons(self, uint8_view):
        """Test comparison operations."""
        # __lt__
        assert uint8_view < 12
        assert not uint8_view < 8

        # __le__
        assert uint8_view <= 12
        assert uint8_view <= 10
        assert not uint8_view <= 4

        # __eq__
        assert uint8_view == 10
        assert 10 == uint8_view  # pylint: disable=misplaced-comparison-constant
        assert not uint8_view == 18
        assert not 18 == uint8_view  # pylint: disable=misplaced-comparison-constant

        # __ne__
        assert uint8_view != 20
        assert 20 != uint8_view  # pylint: disable=misplaced-comparison-constant
        assert not uint8_view != 10
        assert not 10 != uint8_view  # pylint: disable=misplaced-comparison-constant

        # __gt__
        assert uint8_view > 2
        assert not uint8_view > 16

        # __ge__
        assert uint8_view >= 10
        assert uint8_view >= 9
        assert not uint8_view >= 11

    def test_conversions(self, uint8_view):
        """Test type conversion operations."""
        assert str(uint8_view) == "10"
        assert float(uint8_view) == 10.0

    def test_numeric(self, uint8_view):
        """Test comparison operations."""
        # __add__ and __radd__
        assert uint8_view + 2 == 12
        assert 2 + uint8_view == 12

        # __sub__ and __rsub__
        assert uint8_view - 4 == 6
        assert 24 - uint8_view == 14

        # __mul__ and __rmul__
        assert uint8_view * 4 == 40
        assert 6 * uint8_view == 60

        # __truediv__ and __rtruediv__
        assert uint8_view / 2 == 5
        assert 30 / uint8_view == 3

        # __mod__ and __rmod__
        assert uint8_view % 6 == 4
        assert 48 % uint8_view == 8

        # __divmod__ and __rdivmod__
        assert divmod(uint8_view, 3) == (3, 1)
        assert divmod(24, uint8_view) == (2, 4)

        # __pow__ and __rpow__
        assert uint8_view**3 == 1000
        assert 2**uint8_view == 1024

        # __ceil__ and __floor__
        assert ceil(uint8_view) == 10
        assert floor(uint8_view) == 10

        # __floordiv__ and __rfloordiv__
        assert uint8_view // 2 == 5
        assert 20 // uint8_view == 2

        # __round__ and __trunc__
        assert round(uint8_view, 0) == 10
        assert trunc(uint8_view) == 10

    def test_unary(self, uint8_view):
        """Test unary operations."""
        assert -uint8_view == -10  # __neg__
        assert +uint8_view == 10  # __pos__
        assert abs(uint8_view) == 10  # __abs__
        assert int(uint8_view) == 10  # __int__
        assert ~uint8_view == -11  # __invert__

    def test_pack_pos_argument(self, uint8_view):
        """Test pack() utility supports view as positional argument."""
        assert pack(uint8_view, uint8) == bytearray(b"\x0a")

    def test_pack_kw_argument(self, uint8_view):
        """Test pack() utility supports view as keyword argument."""
        assert pack({"x": uint8_view}, {"x": uint8}) == bytearray(b"\x0a")

    def test_unpack(self):
        """Test unpack() returns unpacked value."""
        buffer = bytearray(b"\x01\x02\x0A\x0B")
        value = uint8.view(buffer, offset=2).unpack()
        assert value == 10
        # changing buffer does not effect previously returned value
        buffer[0:4] = [0] * 4
        assert value == 10

    def test_set(self):
        """Test set()."""
        buffer = bytearray(4)
        int_view = uint8.view(buffer, offset=2)
        assert int_view == 0
        int_view.set(1)
        assert int_view == 1
        assert buffer == bytearray([0x00, 0x00, 0x01, 0x00])

    def test_representations(self, uint8_view):
        """Test representations."""
        expected_dump = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            | 2      | 10    | 0a    | uint8  |
            +--------+-------+-------+--------+
            """
        )
        expected_repr = Baseline(
            """
            <uint8 view at 0x2: 10>
            """
        )

        assert str(uint8_view) == "10"
        assert repr(uint8_view) == expected_repr
        assert str(uint8_view.dump) == expected_dump

    def test_dereference(self, uint8_view):
        """Test dereference operation exception."""

        with pytest.raises(RuntimeError) as trap:
            uint8_view[0]  # pylint: disable=pointless-statement

        expected = Baseline(
            """
            <transform 'uint8'> type does not support pointer dereference
            """
        )

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test bytes transform support for structure's sized_member()."""

import pytest
from baseline import Baseline
from plum.bytes import BytesX
from plum.conformance import wrap_message
from plum.exceptions import PackError

bytes4 = BytesX(name="bytes4", nbytes=4)


class Test:
    def test_pack(self):
        expected_message = Baseline(
            """
            +--------+--------+----------+----------------+--------+
            | Offset | Access | Value    | Bytes          | Format |
            +--------+--------+----------+----------------+--------+
            |        |        |          |                | bytes4 |
            | 0      | [0:5]  | b'12345' | 31 32 33 34 35 |        |
            +--------+--------+----------+----------------+--------+

            ValueError occurred during pack operation:

            expected length to be 4 but instead found 5
            """
        )

        with pytest.raises(PackError) as trap:
            bytes4.pack(b"12345")

        assert wrap_message(trap.value) == expected_message
